"""
    IterativeLrfSolver.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data
from SimpleGuinier import SimpleGuinier
from molass_legacy.Trimming.Sigmoid import sigmoid
from Theory.SolidSphere import get_boundary_params

class IterativeLrfSolver:
    def __init__(self, qv, D, E, i=None, denoise=True):
        if i is None:
            from bisect import bisect_right
            i = bisect_right(qv, 0.02)

        self.qv = qv
        self.i = i
        if denoise:
            D_ = get_denoised_data(D, rank=2)
        else:
            D_ = D
        self.D_ = D_
        self.E = E
        cy = np.average(D_[i-5:i+6,:], axis=0)
        self.cy_init = cy
        self.j = np.argmax(cy)
        self.Cinit = np.array([cy, cy**2])

    def solve(self, maxiter=10000, L=1, rg_refresh_cycle=20000, limit_ratio=0.003, debug=False):
        qv = self.qv
        i = self.i
        j = self.j
        cy_init = self.cy_init
        C = self.Cinit
        D_ = self.D_
        pjv = D_[:,j]/cy_init[j]
        E = self.E
        changes = []
        aq_history = []
        last_ratio = None

        for k in range(maxiter+1):
            Cinv = np.linalg.pinv(C)
            P = D_ @ Cinv
            aq, bq = P.T
            c1, c2 = C[:,j]

            if k % rg_refresh_cycle == 0:
                def compute_bounds(rg_only=False):
                    D_pinv = np.linalg.pinv(D_)
                    W = np.dot(D_pinv, P)
                    Pe = np.sqrt(np.dot(E**2, W**2))

                    data = np.array([qv, aq, Pe[:,0]]).T
                    sg = SimpleGuinier(data)
                    Rg = sg.Rg
                    print([k], "Rg=", Rg)
                    if rg_only:
                        return Rg
                    R = np.sqrt(5/3)*Rg
                    bound = 1/(qv*L*R)**2

                    bq_bound = bound * aq/c1
                    bq_bounds = (-bq_bound, bq_bound)
                    return Rg, R, bq_bounds
                Rg, R, bq_bounds = compute_bounds()
                aq_history.append(aq)
                last_ratio = None       # to avoid too early finish

            if k == 0:
                aq_init = aq
                c1_init = c1

            coerced_bq = P[:,1].copy()

            where = bq < bq_bounds[0]
            coerced_bq[where] = bq_bounds[0][where]
            where = bq > bq_bounds[1]
            coerced_bq[where] = bq_bounds[1][where]

            changed_norm_ratio = np.linalg.norm(coerced_bq - P[:,1])/np.linalg.norm(P[:,1])
            if k % 1000 == 0:
                print([k], "changed_norm_ratio=", changed_norm_ratio)
                changes.append(changed_norm_ratio)

            if (k >= maxiter
                or changed_norm_ratio < limit_ratio 
                or last_ratio is not None and changed_norm_ratio > last_ratio
                ):
            # if k >= maxiter:
                print([k], changed_norm_ratio, last_ratio)
                # recompute the filnal Rg
                Rg = compute_bounds(rg_only=True)
                k_final = k
                aq_history.append(aq)
                break


            last_ratio = changed_norm_ratio
            P = np.array([aq + (bq - coerced_bq)*c2/c1, coerced_bq]).T
            # P[:,1] = coerced_bq
            # P = np.array([aq, coerced_bq]).T
            Pinv = np.linalg.pinv(P)
            C = Pinv @ D_

        b1, b2, k_sig = get_boundary_params(Rg, qv=qv)
        w = sigmoid(qv, 1, b1, k_sig, 0)
        aq_init_ = aq_init*c1/c1_init
        modified_aq = aq_init_*(1 - w) + aq*w
        print("c1_init, c1=", c1_init, c1)

        if debug:
            glim_i = bisect_right(qv, 2.0/Rg)
            gslice = slice(0, glim_i)
            x_ = np.arange(len(changes))*1000
            qv2 = qv[gslice]**2

            with plt.Dp():
                fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
                ax1.set_title("Norm Difference Ratio Changes")
                ax1.plot(x_, changes)
                ax2.set_title("A(q) Changes")
                ax3.set_title("A(q) Changes (Guinier Plot)")
                ax2.set_yscale("log")
                ax2.plot(qv, pjv, label="data")
                ax3.plot(qv2, np.log(pjv[gslice]), label="data")
                for k_, aq_ in enumerate(aq_history):
                    label_k = k_*rg_refresh_cycle if k_ < len(aq_history)-1 else k_final
                    label = "[%d] A(q)" % label_k
                    ax2.plot(qv, aq_, label=label)
                    ax3.plot(qv2, np.log(aq_[gslice]), label=label)
                ax2.plot(qv, modified_aq, ":", color="red", lw=2, label="modifeid A(q)")
                ax3.plot(qv2, np.log(modified_aq[gslice]), ":", color="red", lw=2, label="modifeid A(q)")
                ax2.legend()
                ax3.legend()
                axt = ax2.twinx()
                axt.grid(False)
                axt.plot(qv, w, ":", color="cyan")
                fig.tight_layout()
                plt.show()

        P[:,0] = modified_aq

        return P, C, Rg, R, L, bq_bounds, coerced_bq, w, np.array(changes)



"""
    RgProcess.RgCurve.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.SliceUtils import slice_consecutives
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Trimming import TrimmingInfo   # used in eval_file

def compute_rg_segment(qv, D, E, slice_, rg_buffer, qu_buffer, progress_cb=None):
    from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

    includes_nan = False
    for j in range(slice_.start, slice_.stop):
        data = np.array([qv, D[:,j], E[:,j]]).T
        sg = SimpleGuinier(data)
        if sg.Rg is None:
            includes_nan = True
        rg = sg.Rg
        rg_buffer[j] = rg
        qu_buffer[j] = sg.basic_quality
        if progress_cb is not None:
            progress_cb(rg_buffer, j)

    if includes_nan:
        # as in 20171226
        import logging
        logger = logging.getLogger(__name__)
        buffer = rg_buffer[slice_]
        pad_value = np.mean(buffer[np.logical_not(np.isnan(buffer))])
        where = slice_.start + np.where(np.isnan(buffer))[0]
        buffer[np.isnan(buffer)] = pad_value
        logger.warning("NaN Rg's have been replaced by mean value %g of non-NaN values at %s.", pad_value, str(where))

def compute_rg_segment_proxy(slice_, rg_buffer, qu_buffer, progress_cb=None):
    if progress_cb is None:
        return

    for j in range(slice_.start, slice_.stop):
        progress_cb(rg_buffer, j)

def compute_rg_segment_wrapper(arg):
    from molass_legacy.KekLib.SharedArrays import SharedArrays
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.DataStructure.MatrixData import simple_plot_3d
    name, tuples, slice_ = arg
    print("name=", name, "tuples=", tuples)
    sa = SharedArrays(name=name, tuples=tuples)
    qv, D, E = sa.get_arrays()
    rg_buffer = np.zeros(D.shape[1])
    compute_rg_segment(qv, D, E, slice_, rg_buffer)
    if False:
        print(slice_)
        plt.push()
        fig = plt.figure(figsize=(14,6))
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122)
        simple_plot_3d(ax1, D)
        ax2.plot(rg_buffer)
        plt.show()
        plt.pop()
    return rg_buffer

def make_availability_slices(y, max_y=None, min_ratio=0.03, min_size=20):
    if max_y is None:
        max_y = np.max(y)

    # int(f), int(t) is meant to convert np.int64 to int in 64GB or larger RAM environment
    # to avoid str(slice_) to become like 'slice(np.int64(202), np.int64(258), None)'
    pairs = [(int(f), int(t)) for f, t in slice_consecutives(np.where(y/max_y >= min_ratio)[0])]

    slices = []
    states = []
    start = 0
    for f, t in pairs:
        size = t - f
        if size < min_size:
            continue

        if start < f:
            slices.append(slice(start, f))
            states.append(0)

        start = t+1
        slices.append(slice(f, start))
        states.append(1)
    if start < len(y):
        slices.append(slice(start, len(y)))
        states.append(0)
    return slices, states

def rg_folder_ok_stamp_path(rg_folder):
    return os.path.join(rg_folder, 'ok.stamp')

def check_rg_folder(rg_folder):
    ret_state = os.path.exists(rg_folder) and os.path.exists(rg_folder_ok_stamp_path(rg_folder))

    if ret_state:
        baseline_type = get_setting("unified_baseline_type")
        baseline_type_txt = os.path.join(rg_folder, "baseline_type.txt")
        if os.path.exists(baseline_type_txt):
            from molass_legacy.KekLib.EvalUtils import eval_file
            prev_baseline_type = eval_file(baseline_type_txt, locals_=globals())
        else:
            prev_baseline_type = None
        ret_state = prev_baseline_type == baseline_type

    return ret_state

class RgCurve:
    def __init__(self, qv, ecurve, D, E, xy_info=None, min_ratio=0.03, multi_process=False, progress_cb=None):
        from time import time
        import logging
        self.logger = logging.getLogger(__name__)
        self.progress_cb = progress_cb

        t0 = time()
        if xy_info is None:
            x = ecurve.x
            y = ecurve.y
            max_y = ecurve.max_y
        else:
            x, y, max_y = xy_info

        slices, states = make_availability_slices(y, max_y=max_y, min_ratio=min_ratio)

        if multi_process:
            # not tested
            segments, qualities = self.create_segments_multi_process(qv, x, y, D, E, slices, states)
        else:
            segments, qualities = self.create_segments_single_process(qv, x, y, D, E, slices, states)

        self.x = x
        self.y = y
        self.ecurve = ecurve
        self.slices = slices
        self.states = states
        self.segments = segments
        self.qualities = qualities
        xr_restrict_list = get_setting("xr_restrict_list")
        self.rg_trimming = None if xr_restrict_list is None else xr_restrict_list[0]
        self.baseline_type = get_setting("unified_baseline_type")
        self.X = None
        self.excl_info = None
        self.excl_spline = None
        self.logger.info("It took %.3g seconds for rg_curve construction.", time()-t0)

    def get_valid_slices(self):
        return self.slices

    def get_mask(self):
        mask = np.zeros(len(self.x), dtype=bool)
        for slice_, state in zip(self.slices, self.states):
            if state:
                mask[slice_] = True
        return mask

    def get_weights(self):
        weight_list = []
        k = 0
        for slice_, state in zip(self.slices, self.states):
            if state:
                if self.qualities is None:
                    qu = None
                else:
                    qu = self.qualities[k]
                if qu is None:
                    qu = np.ones(slice_.stop - slice_.start)
                weight_list.append(qu)
                k += 1
        return np.concatenate(weight_list)

    def create_segments_multi_process(self, qv, x, y, D, E, slices, states):
        from molass_legacy.KekLib.SharedArrays import SharedArrays
        from molass_legacy.KekLib.SciPyCookbook import smooth
        from multiprocessing import Pool

        sa = SharedArrays([qv, D, E])
        sa_info = (sa.name, sa.get_tuples())

        args_list = []
        arg_slices = []
        for slice_, state in zip(slices, states):
            if state == 0:
                continue
            args_list.append((*sa_info, slice_))
            arg_slices.append(slice_)

        with Pool(2) as p:
            ret_list = p.map(compute_rg_segment_wrapper, args_list)

        segments = []
        qualities = []
        for ret, slice_ in zip(ret_list, arg_slices):
            segments.append((x[slice_], y[slice_], smooth(ret[slice_])))
            # qu = 
            # qualities.append(qu)
        return segments, qualities

    def create_segments_single_process(self, qv, x, y, D, E, slices, states):
        from molass_legacy.KekLib.SciPyCookbook import smooth

        rg_buffer_file = get_setting("rg_buffer_file")
        if rg_buffer_file is None:
            rg_buffer = np.zeros(len(x))
            qu_buffer = np.zeros(len(x))
        else:
            rg_buffer = np.loadtxt(rg_buffer_file)
            qu_buffer = None

        segments = []
        qualities = []
        for slice_, state in zip(slices, states):
            if state == 0:
                continue
            if rg_buffer_file is None:
                compute_rg_segment(qv, D, E, slice_, rg_buffer, qu_buffer, progress_cb=self.progress_cb)
                rg = smooth(rg_buffer[slice_])
                qu = qu_buffer[slice_]
            else:
                compute_rg_segment_proxy(slice_, rg_buffer, qu_buffer, progress_cb=self.progress_cb)
                rg = rg_buffer[slice_]
                qu = None
            segments.append((x[slice_], y[slice_], rg))
            qualities.append(qu)
        return segments, qualities

    def get_curve_segments(self):
        return self.segments

    def get_valid_curves(self):
        x_segs = []
        y_segs = []
        rg_segs = []
        for x, y, rg in self.segments:
            x_segs.append(x)
            y_segs.append(y)
            rg_segs.append(rg)
        return [np.concatenate(segs) for segs in [x_segs, y_segs, rg_segs]]

    def add_exclspline(self, poresize=70, return_rhov=False, return_excl_xy=False):
        from scipy.optimize import minimize
        from scipy.interpolate import UnivariateSpline
        x = self.ecurve.x
        segments = self.get_curve_segments()
        k = 0
        tr_list = []
        rg_list = []
        qu_list = []
        for x_, y_, rg in segments:
            tr_list.append(x_)
            rg_list.append(rg)
            qu = self.qualities[k]
            qu_list.append(qu)
            k += 1
        trv = np.concatenate(tr_list)
        rgv = np.concatenate(rg_list)
        quv = np.concatenate(qu_list)
        rhov = rgv/poresize
        rhov[rhov > 1] = 1
        # print("rhov=", rhov)
        if return_rhov:
            return rhov

        def objective(p):
            t0, K = p
            trv_ = t0 + K*(1 - rhov)**3
            return np.sum(quv * (trv_ - trv)**2)

        res = minimize(objective, (0, 1000))
        t0, K = res.x
        excl_x = t0 + K*(1 - rhov)**3
        excl_y = rhov*poresize
        if return_excl_xy:
            return excl_x, excl_y

        x_ = np.flip(excl_x)
        y_ = np.flip(excl_y)
        ux_, uy_ = np.unique(np.array([x_, y_]), axis=1)        

        self.excl_info = poresize, t0, K
        self.excl_spline = UnivariateSpline(ux_, uy_)

    def get_probabilistic_data(self):
        if self.X is None:
            from Rgg.RggUtils import convert_to_probabilitic_data
            X_list = []
            for x, y, rg in self.get_curve_segments():
                X_list.append(convert_to_probabilitic_data(x, y, rg, max_y =self.ecurve.max_y))
            self.X = np.concatenate(X_list)
        return self.X

    def proof_plot_impl(self, fig, axes):
        import molass_legacy.KekLib.MatplotlibUtils      # required for annotate3D
        from Rgg.RggUtils import plot_histogram_2d

        ax1, ax2, ax3 = axes
        ecurve = self.ecurve
        segments = self.get_curve_segments()
        x = ecurve.x
        y = ecurve.y
        max_y = ecurve.max_y
        X = self.get_probabilistic_data()
        fig.suptitle("Generation of 2D Probabilistic Data", fontsize=20)
        ax1.set_title("Xray Elution and Smoothed Rg Curve", fontsize=16)
        ax2.set_title("Rg Qualities", fontsize=16)
        ax3.set_title("Histogram of the 2D Propbabilistic Data", fontsize=16)
        axt = ax1.twinx()
        axt.grid(False)

        ax1.plot(x, y, color='orange')
        ax2.set_xlim(ax1.get_xlim())

        k = 0
        for x_, y_, rg in segments:
            axt.plot(x_, rg, ':', color='C1')
            qu = self.qualities[k]
            ax2.plot(x_, qu)
            plot_histogram_2d(ax3, x_, y_, rg, max_y)
            k += 1

        ymin, ymax = axt.get_ylim()
        axt.set_ylim(min(10, ymin), max(50, ymax))

        ptx = ecurve.get_primarypeak_x()
        pti = ecurve.get_primarypeak_i()
        k = 0
        for slice_, state in zip(self.slices, self.states):
            if state == 0:
                continue
            if slice_.start <= pti and pti < slice_.stop:
                pty = segments[k][2][pti - slice_.start]
                break
            k += 1

        ptz = 100
        peaktop = (ptx, pty, ptz)
        ax3.annotate3D('(%.3g, %.3g, %.3g)' % peaktop, peaktop,
              xytext=(30,-30),
              textcoords='offset points',
              bbox=dict(boxstyle="round", fc="lightyellow"),
              arrowprops = dict(arrowstyle="-|>", ec='black', fc='white', lw=1))

        ax3.set_xlabel('Eno')
        ax3.set_ylabel('Rg')
        ax3.set_zlabel('Counts')

        ax3.set_xlim(ax1.get_xlim())
        ax3.set_ylim(10, 60)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)

    def proof_plot(self):
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            figsize=(21,7)
            fig = plt.figure(figsize=figsize)
            ax1 = fig.add_subplot(131)
            ax2 = fig.add_subplot(132)
            ax3 = fig.add_subplot(133, projection='3d')
            self.proof_plot_impl(fig, (ax1, ax2, ax3))
            plt.show()

    def plot(self, fig=None, ax=None):
        if fig is None:
            import matplotlib.pyplot as plt
            fig = plt.figure()
        if ax is None:
            ax = fig.add_subplot()

        axt = ax.twinx()
        axt.grid(False)

        ax.set_title("$R_g$ Curve Object Plot")

        ecurve = self.ecurve
        segments = self.get_curve_segments()
        x = ecurve.x
        y = ecurve.y
        ax.plot(x, y, label='data')
        k = 0
        for x_, y_, rg in segments:
            label = 'Guinier $R_g$ curve' if k == 0 else None
            axt.plot(x_, rg, ':', color='gray', label=label)
            k += 1

        if self.excl_spline is not None:
            poresize, t0 = self.excl_info[0:2]
            ex = x[x > t0]
            ey = self.excl_spline(ex)
            axt.plot(ex, ey, color='yellow', label='SEC $R_g$ curve')
            axt.axvline(x=t0, linestyle=':', color='red', label='$t_0$ (poresize=%.3g)' % poresize)

        axt.legend()
        fig.get_tight_layout()

    def save_buffer(self, path):
        buffer = np.zeros(len(self.x))
        k = 0
        for slice_, state in zip(self.slices, self.states):
            if state == 0:
                continue
            buffer[slice_] = self.segments[k][2]
            k += 1
        np.savetxt(path, buffer)

    def export(self, folder):
        import os

        precision_save = np.get_printoptions()["precision"]
        np.set_printoptions(precision=17)

        for filename, item in [
                ("slices.txt", self.slices),
                ("states.txt", self.states),
                ("segments.txt", self.segments),
                ("qualities.txt", self.qualities),
                ("rg_trimming.txt", self.rg_trimming),
                ("baseline_type.txt", self.baseline_type),
                ]:
            with open(os.path.join(folder, filename), "w") as fh:
                fh.write(str(item))

        with open(rg_folder_ok_stamp_path(folder), "w") as fh:
            fh.write("")

        np.set_printoptions(precision=precision_save)

    def get_rgs_from_trs(self, trs):
        if self.excl_spline is None:
            self.add_exclspline()
        return self.excl_spline(trs)

    def get_rgs_from_trs_depricated(self, trs):
        from bisect import bisect_right
        ret_rgs = np.ones(len(trs)) * np.nan
        for k, tr in enumerate(trs):
            for x, y, rg in self.segments:
                if x[0] <= tr and tr <= x[-1]:
                    j = bisect_right(x, tr)
                    ret_rgs[k] = rg[j]
                    break
        return ret_rgs

RG_CALLBACK_CYCLE = 10

class ProgressCallback:
    def __init__(self, queue, f, t):
        self.queue = queue
        self.f = f
        self.t = t

    def __call__(self, rg_buffer, j):
        if j % RG_CALLBACK_CYCLE == 0:
            rate = j/len(rg_buffer)
            progress = self.f*(1 - rate) + self.t*rate
            self.queue.put([progress, (rg_buffer, j)])

def draw_rg_bufer(ax, p_info, dialog, x):
    rg_buffer, j = p_info
    draw = j % RG_CALLBACK_CYCLE == 0
    if draw:
        # drawing each seems too heavy
        rg_positive = rg_buffer > 0
        x = x[rg_positive]
        y = rg_buffer[rg_positive]
        if dialog.rg_line is None:
            dialog.rg_line, = ax.plot(x, y, color='gray', alpha=0.5, label='Rg')
            ax.legend(loc='center left')
        else:
            dialog.rg_line.set_data(x, y)
        ax.set_ylim(10, np.max(y))
        print([j], "drawn")

    return draw

def compute_init_rgs(rg_curve, n_components):
    x_, y_, rg_ = rg_curve.get_valid_curves()
    mean = np.mean(rg_)
    std = np.std(rg_)
    init_rgs = np.linspace(mean+std, mean-std, n_components)
    return init_rgs



# coding: utf-8
"""

    ファイル名：   CanvasFrame.py

    処理内容：

        デバッグ用の plot 

    Copyright (c) 2017-2021, Masatsuyo Takahashi, KEK-PF

"""
import matplotlib.pyplot                as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk
import matplotlib

class CanvasFrame(Tk.Frame):
    def __init__( self, parent, figsize=None, func=None ):
        Tk.Frame.__init__( self, parent )

        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'
        cframe = Tk.Frame( self )
        cframe.pack()

        self.fig = fig = plt.figure( figsize=figsize )
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )

        if func is not None:
            func( fig )
        self.show()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

    def close_fig(self):
        plt.close(self.fig)

    def draw( self, func ):
        func( self.fig )
        self.show()

    def show( self ):
        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()


"""
    DecompDummyFrame.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk

class DecompDummyFrame(Tk.Frame):
    def __init__(self, parent, model):
        self.model = model
        Tk.Frame.__init__(self, parent)

    def is_delayed(self):
        return self.model.is_delayed()

    def get_decomp_score(self):
        # return a very large value to avoid being selected
        return 1e+10

    def get_fit_error(self):
        # return a very large value to avoid contradiction
        return 1e+10

    def close_figs(self):
        # nothing to do
        pass



"""
    Estimators.SdmEstimatorProxy.py

    temporary fix to make the get_colparam_bounds available

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import os
import numpy as np
from .SdmEstimator import SdmEstimator
from molass_legacy.Optimizer.TheUtils import FILES

class SdmEstimatorProxy(SdmEstimator):
    def __init__(self, jobfolder):
        bounds_text_path = os.path.join(jobfolder, FILES[7])    # "bounds.txt"
        self.bounds = np.loadtxt(bounds_text_path)


"""
    GradualRegionAnalysis.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import re
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from ScrolledFrame import ScrolledFrame
from DataUtils import get_in_folder
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy._MOLASS.SerialSettings import get_setting

class GradualRegionAnalysis(Dialog):
    def __init__(self, parent, dialog, sd, trimming_info, pre_recog, debug=False):
        self.logger = logging.getLogger(__name__)
        print("pre_recog=", pre_recog)

        D, E, qv, xr_ecurve = sd.get_xr_data_separate_ly()

        self.qv = qv
        i_smp = sd.xray_index
        eslice, m = self.determine_analysis_range(xr_ecurve)
        self.i_smp = i_smp
        self.y = D[:,m]
        cv = pre_recog.cs.get_uniformly_mapped_a_curve()

        if debug:
            import molass_legacy.KekLib.DebugPlot as dplt
            x = xr_ecurve.x
            y = xr_ecurve.y
            with dplt.Dp():
                fig, ax = dplt.subplots()
                ax.set_title("cv debug")
                ax.plot(x, y)
                axt = ax.twinx()
                axt.grid(False)
                axt.plot(x, cv)
                fig.tight_layout()
                dplt.show()

        self.eslice = eslice
        self.aslice = trimming_info.get_slice()
        self.info_list = self.compute_info_with_changed_trimming(D, E, qv, i_smp, eslice, self.aslice, cv)

        Dialog.__init__(self, parent, "Gradual Region Analysis", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):

        in_folder = get_in_folder()
        suptitle = Tk.Label(body_frame, text="Guinier Analysis Results with gradually changed trimming on %s" % in_folder, font=('', 18))
        suptitle.pack(pady=10)

        cframe1 = Tk.Frame(body_frame)
        cframe1.pack(side=Tk.LEFT)
        space_width = 20
        space = Tk.Frame(body_frame, width=space_width)
        space.pack(side=Tk.LEFT)

        cframe2 = Tk.Frame(body_frame)
        cframe2.pack(side=Tk.LEFT, fill=Tk.X, expand=1)

        cframe21 = Tk.Frame(cframe2, bg="white")
        cframe21.pack(fill=Tk.X, expand=1)
        cframe22 = Tk.Frame(cframe2)
        cframe22.pack()
        cframe2_ = ScrolledFrame(cframe22, bg="white")
        cframe2_.pack()
        self.scrolled_frame = cframe2_

        fig1, ax1 = plt.subplots(figsize=(6,8))
        self.fig1 = fig1
        self.ax1 = ax1

        qv = self.qv
        n = len(self.info_list)

        data_title = "" if get_setting('test_pattern') is None else " in " + in_folder

        ax1.set_title("$R_g$ Changes" + data_title, fontsize=16)
        ax1.set_ylabel("$R_g$")
        ax1.set_xlabel("trimming start Q")
        rg_list = [rec[1].Rg for rec in self.info_list]
        ax1.plot(qv[0:n], rg_list, "-o")

        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)
        tq = self.qv[self.aslice.start]
        ax1.plot([tq, tq], [ymin, ymax], color="yellow")

        fig1.tight_layout()

        self.mpl_canvas1 = FigureCanvasTkAgg(fig1, cframe1)
        self.mpl_canvas_widget1 = self.mpl_canvas1.get_tk_widget()
        self.mpl_canvas_widget1.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas1.mpl_connect('button_press_event', self.on_click)

        label = Tk.Label(cframe21, text="Guinier Plots", font=('Arial', 14), bg="white")
        label.pack(fill=Tk.X, expand=1, pady=5)

        nrows = len(self.info_list)
        height = 3*nrows
        fig2 = plt.figure(figsize=(11,height))
        self.fig2 = fig2

        qv2 = self.qv**2
        gstop = np.max([rec[1].guinier_stop for rec in self.info_list])

        gs = GridSpec(nrows, 5)
        for i, (pv, sg) in enumerate(self.info_list):
            ax0 = fig2.add_subplot(gs[i,0])
            ax0.set_xticks([])
            ax0.set_yticks([])
            ax0.set_facecolor("white")
            ax0.text(0.5, 0.5, "Q[%d]\n%.3g\n\n$R_q$=%.3g" % (i, qv[i], sg.Rg), fontsize=16, ha="center", va="center", alpha=0.5)

            ax1 = fig2.add_subplot(gs[i,1:])
            ax1.set_xlabel("$Q^2$")
            ax1.set_ylabel("$ln(I)$")

            x2 = qv2[i:gstop]
            if i == 0:
                scaled_y = self.y[:gstop] * pv[self.i_smp]/self.y[self.i_smp]
            ax1.plot(qv2[:gstop], np.log(scaled_y), "o", markersize=3, label="measured raw data")
            ax1.plot(x2, np.log(pv[:len(x2)]), "o", color="C1", markersize=3, label="LRF-corrected data")
            ax1.plot(sg.guinier_x, sg.guinier_y, color="red", alpha=0.5, label="estimated guinier interval")
            ax1.legend()

            if i == 0:
                xmin, xmax = ax1.get_xlim()
            else:
                ax1.set_xlim(xmin, xmax)

        fig2.tight_layout()
        self.mpl_canvas2 = FigureCanvasTkAgg(fig2, cframe2_.interior)
        self.mpl_canvas_widget2 = self.mpl_canvas2.get_tk_widget()
        self.mpl_canvas_widget2.pack(fill=Tk.BOTH, expand=1)

        self.set_geometry(space_width)
        self.update()

        self.buttonbox(build=True)

    def buttonbox(self, build=False):
        if not build:
            # temporary bug fix
            # ignore this standared callback
            return

        box = Tk.Frame(self)
        box.pack(side=Tk.BOTTOM)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def set_geometry(self, space_width):
        self.update()
        canvas_width1 = int(self.mpl_canvas_widget1.cget('width'))
        canvas_width2 = int(self.mpl_canvas_widget2.cget('width'))

        geometry = self.geometry()
        print("geometry=", geometry, canvas_width1, canvas_width2)
        width = canvas_width1 + space_width + canvas_width2 + 40
        height = int(self.mpl_canvas_widget1.cget('height')) + 60

        new_geometry = re.sub( r'(\d+x\d+)(.+)', lambda m: "%dx%d" % (width, height) + m.group(2), geometry)
        print("new_geometry=", new_geometry)
        self.geometry(new_geometry)

    def determine_analysis_range(self, xr_ecurve):
        eranges = xr_ecurve.get_default_editor_ranges()
        print("eranges=", eranges)
        pno = xr_ecurve.primary_peak_no
        range_ = eranges[pno]
        f = range_[0][0]
        m = range_[0][1]
        t = range_[-1][-1]
        return slice(f, t+1), m

    def compute_info_with_changed_trimming(self, D, E, qv, i_smp, eslice, aslice, cv):
        cv_ = cv[eslice]
        C = np.array([cv_, cv_**2])
        Cinv = np.linalg.pinv(C)

        info_list = []
        for i in range(i_smp):
            aslice_ = slice(i, aslice.stop)
            qv_ = qv[aslice_]
            D_ = D[aslice_,eslice]
            E_ = E[aslice_,eslice]
            P = D_@Cinv
            Dinv = np.linalg.pinv(D_)
            W   = np.dot(Dinv, P)
            Pe  = np.sqrt(np.dot(E_**2, W**2))
            data = np.array([qv_, P[:,0], Pe[:,0]]).T
            sg = SimpleGuinier(data)
            # print([i], sg.Rg)
            info_list.append((P[:,0], sg))

        return info_list

    def on_click(self, event):
        if event.inaxes != self.ax1:
            return

        q = event.xdata
        qv_ = self.qv[:self.i_smp]
        k = np.argmin((qv_ - q)**2)
        self.scrolled_frame.canvas.yview_moveto(k/len(self.info_list))

    def cancel(self):
        # overiding cancel to cleanup self.fig1 and self.fig2
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig1)
        plt.close(self.fig2)
        Dialog.cancel(self)

def survey_summary(survey_csv):
    from scipy.stats import linregress
    data_list = []
    with open(survey_csv) as fh:
        for line in fh:
            fields = line[:-1].split(',')
            values = np.array([float(v) for v in fields[2:]])
            n = int(values[0])
            qv = values[1:1+n]
            rgv = values[1+n:1+2*n]
            # print(qv)
            # print(rgv)
            slope, intercept, r_value, p_value, std_err = linregress(qv, rgv)
            if std_err > 100:
                print("skipping", fields[1], r_value, std_err)
                continue
            print(fields[1], r_value, std_err)
            data_list.append((r_value, std_err))
    data = np.array(data_list)

    fig, ax = plt.subplots()
    ax.set_title("Scatter Plot of Rg Variations in Gradual Region Analysis", fontsize=16)
    ax.plot(data[:,0], data[:,1], "o", markersize=3)

    ax.set_xlim(-1, 1)

    ax.set_xlabel("r_value", fontsize=16)
    ax.set_ylabel("std_error", fontsize=16)
    fig.tight_layout()
    plt.show()



# coding: utf-8
"""
    Pomegranate.MappingGmmDecomposer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from itertools import combinations
from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from molass_legacy.Peaks.RobustPeaks import RobustPeaks
from molass_legacy.Peaks.ElutionModels import egh
from Prob.GaussianMixture import hist_to_source
from Prob.EghMixture import EghMixture
import OurStatsModels as sm

APPLY_PARAM_CONSTRAINTS = True
USE_WLS = True

class MappingGmmDecomposer:
    def __init__(self, uv_rp, xr_rp, num_components=None):
        self.uv_rp = uv_rp
        self.xr_rp = xr_rp
        self.uv_x = uv_rp.x
        self.uv_y = uv_rp.y
        self.xr_x = xr_rp.x
        self.xr_y = xr_rp.y
        self.num_components = num_components

    def decompose_separatley(self):
        from .EghMixtureModel import EghMixtureModel

        num_components = self.num_components
        x0 = self.uv_x
        y0 = self.uv_y
        mm0 = EghMixtureModel(x0, y0, num_components)
        x1 = self.xr_x
        y1 = self.xr_y
        mm1 = EghMixtureModel(x1, y1, num_components)

        best_pair = None
        best_r_value = None
        last_max_r_value = None
        for k in range(5):
            results0 = []
            if best_pair is None or True:
                res0, res1 = None, None
            else:
                res0, res1 = best_pair
            # print([k], res0, res1)
            print([k], "----")
            for i in range(2):
                results0.append(mm0.fit(init_result=res0))

            results1 = []
            for i in range(2):
                results1.append(mm1.fit(init_result=res1))

            max_pair, max_r_value, _ = self.get_best_combination(results0, results1)
            if last_max_r_value is not None and last_max_r_value > 0.95:
                if max_r_value < last_max_r_value:
                    pass

            if best_r_value is None or max_r_value > best_r_value:
                best_r_value = max_r_value
                best_pair = max_pair

            last_max_r_value = max_r_value

        res0, res1 = best_pair

        cy_list0, ty0 = mm0.get_components(res0, x0, y0)
        cy_list1, ty1 = mm1.get_components(res1, x1, y1)

        plot_mapped_states([
            [x0, y0, cy_list0, ty0],
            [x1, y1, cy_list1, ty1],
            ])

    def get_best_combination(self, results0, results1):

        max_r_value = None
        max_pair = None
        r_values = []
        for i, res0 in enumerate(results0):
            x = [params[0] for params in res0.params_list]
            for j, res1 in enumerate(results1):
                y = [params[0] for params in res1.params_list]
                r_value = linregress(x, y)[2]
                print((i,j), r_value)
                if max_r_value is None or r_value > max_r_value:
                    max_r_value = r_value
                    max_pair = (i, j)
                    r_values.append(r_value)

        i, j = max_pair
        return (results0[i], results1[j]), max_r_value, np.average(r_values)

def linear_regression(x, y, weights=None):
    if weights is None:
        slope, intercept, r_value, p_value, std_err = linregress(x, y)
    else:
        X = sm.add_constant(x)
        mod = sm.WLS(y, X, weights=weights)
        res = mod.fit()
        intercept, slope = res.params
    return slope, intercept

def plot_mapped_states(components_list, scale=False, step=None, mappinn_info=None):
    plt.push()
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))

    step_str = "" if step is None else " step %d" % step
    fig.suptitle("UV-Xray Decomposed Mapping Figures (Plan)%s" % step_str, fontsize=30)
    ax1.set_title("UV Elution", fontsize=20)
    ax2.set_title("Xray Elution", fontsize=20)
    ax3.set_title("Mapped Elution", fontsize=20)

    for ax, components, color in zip([ax1, ax2], components_list, ['C0', 'C1']):
        x, y, cy_list, ty = components
        if scale:
            s = np.max(y)/np.max(ty)
        else:
            s = 1

        ax.plot(x, y, color=color)
        ax.plot(x, s*ty, ':', color='red')
        for cy in cy_list:
            ax.plot(x, s*cy, ':')

    if mappinn_info is not None:
        mapping_params, uv_params, uv_pi, xr_params, xr_pi = mappinn_info
        slope, intercept = mapping_params
        x0, y0, cy_list0, ty0 = components_list[0]
        x1, y1, cy_list1, ty1 = components_list[1]
        ax3.plot(x1, y1, color='C1', label='Xray')
        spline = UnivariateSpline(x0, y0, s=0, ext=3)
        x_ = x1*slope + intercept
        y_ = spline(x_)
        scale = np.max(y1)/np.max(y_)
        my = y_*scale
        ax3.plot(x1, my, color='gray', alpha=0.2, label='simply mapped UV')
        smy = make_mapped_uv(x1, my, uv_params, uv_pi, xr_params, xr_pi)
        ax3.plot(x1, smy, ':', color='C0', label='scaling mapped UV')
        ax3.legend(fontsize=16, loc="upper left")

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()
    plt.pop()

def make_mapped_uv(x, mapped_y, uv_params, uv_pi, xr_params, xr_pi):
    w = np.zeros(len(mapped_y))
    for uv_p, xr_p, params in zip(uv_pi, xr_pi, xr_params):
        w += egh(x, xr_p/uv_p, *params)
    smy = mapped_y*w
    scale = np.max(mapped_y)/np.max(smy)
    return smy*scale

def spike_demo():
    x0 = np.arange(1000)
    y0 = np.zeros(len(x0))
    for params in [
            [1.0, 401, 50, 5],
            [0.7, 602, 60, 10],
            ]:
        y0 += egh(x0, *params)

    x1 = np.arange(500)
    y1 = np.zeros(len(x1))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y1 += egh(x1, *params)

    spike_demo_impl(x0, y0, x1, y1, num_components=2)

def spike_demo_impl(x0, y0, x1, y1, num_components=None):
    rb0 = RobustPeaks(x0, y0)
    rb1 = RobustPeaks(x1, y1)
    md = MappingGmmDecomposer(rb0, rb1, num_components=num_components)
    md.decompose_separatley()
    # md.decompose()


def spike_demo_real(in_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()
    spike_demo_impl(uv_curve.x, uv_curve.y, xr_curve.x, xr_curve.y, num_components=4)



"""
    Models.Stochastic.LnporeStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from matplotlib.widgets import Slider, Button
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
from SecTheory.BasicModels import robust_single_pore_pdf
from importlib import reload
import Models.Stochastic.LognormalPoreFunc
reload(Models.Stochastic.LognormalPoreFunc)
from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func, lognormal_pore_func
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma, compute_mode, compute_stdev
from molass_legacy.Models.Stochastic.SecModelUtils import DEFAULT_SIGMA, SIGMA_LOWER_BOUND
from molass_legacy.Models.Stochastic.ParamLimits import MAX_PORESIZE

def compute_curves_from_monopore_params(x, params, rgs):
    N, T, x0, me, mp, poresize = params[0:6]
    scales = params[6:]
    cy_list = []
    for k, (rg, scale) in enumerate(zip(rgs, scales)):
        rho = min(1, rg/poresize)
        ni_ = N * (1 - rho)**me
        ti_ = T * (1 - rho)**mp
        cy = scale*robust_single_pore_pdf(x - x0, ni_, ti_)
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    return cy_list, ty, poresize

def compute_curves_from_lnpore_params(x, params, rgs):
    N, T, x0, me, mp, poresize = params[0:6]
    mu, sigma = compute_mu_sigma(poresize, poresize*0.2)
    sigma = max(sigma, SIGMA_LOWER_BOUND)
    scales = params[6:]
    cy_list = []
    for k, (rg, scale) in enumerate(zip(rgs, scales)):
        cy = lognormal_pore_func(x, scale, N, T, me, mp, mu, sigma, rg, x0)
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)
    return cy_list, ty, mu, sigma

def study(x, y, baseline, model, peaks, peak_rgs, props):
    from importlib import reload
    import Models.Stochastic.RoughGuess
    reload(Models.Stochastic.RoughGuess)
    import Models.Stochastic.MonoporeGuess
    reload(Models.Stochastic.MonoporeGuess)
    from molass_legacy.Models.Stochastic.RoughGuess import guess_monopore_params_roughtly
    from molass_legacy.Models.Stochastic.MonoporeGuess import guess_monopore_params_using_moments

    moments_list = compute_egh_moments(peaks)
    rough_monopore_params = guess_monopore_params_roughtly(x, y, model, peaks, peak_rgs, props, moments_list, debug=False)
    print("monopore_params = ", rough_monopore_params)
    if rough_monopore_params is None:
        return
 
    better_monopore_params = guess_monopore_params_using_moments(x, y, moments_list, peak_rgs, props, rough_monopore_params)
    print("better_monopore_params=", better_monopore_params)
    if better_monopore_params is None:
        return

    def plot_elution_curves(ax, title, mnp_cy_list, mnp_ty, lnp_cy_list, lnp_ty):
        ax.set_title(title, fontsize=16)
        ax.plot(x, y, label="data")
        for k, (mnp_cy, lnp_cy) in enumerate(zip(mnp_cy_list, lnp_cy_list)):
            color = "C%d" % (k+1)
            ax.plot(x, mnp_cy, ":", color=color)
            ax.plot(x, lnp_cy, "-", color=color)
        ax.plot(x, mnp_ty, ":", color="red")
        ax.plot(x, lnp_ty, "-", color="red")

    rv = np.arange(800)
    def plot_psd(ax, title, poresize, mu, sigma):
        ax.set_title(title, fontsize=16)
        ax.plot(rv, distr_func(rv, mu, sigma))
        ax.axvline(poresize, color="green")

    mnp_cy_list1, mnp_ty1, poresize1 = compute_curves_from_monopore_params(x, rough_monopore_params, peak_rgs)
    lnp_cy_list1, lnp_ty1, mu1, sigma1 = compute_curves_from_lnpore_params(x, rough_monopore_params, peak_rgs)
    mnp_cy_list2, mnp_ty2, poresize2 = compute_curves_from_monopore_params(x, better_monopore_params, peak_rgs)
    lnp_cy_list2, lnp_ty2, mu2, sigma2 = compute_curves_from_lnpore_params(x, better_monopore_params, peak_rgs)

    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,10))
        fig.suptitle("Lognormal Pore Model Study with Differenent Monopore Params", fontsize=20)
        ax1, ax2 = axes[0,:]
        plot_elution_curves(ax1, "Rough Monopore Elution", mnp_cy_list1, mnp_ty1, lnp_cy_list1, lnp_ty1)
        plot_elution_curves(ax2, "Better Monopore Elution", mnp_cy_list2, mnp_ty2, lnp_cy_list2, lnp_ty2)
        ax3, ax4 = axes[1,:]
        plot_psd(ax3, "Rough Monopore PSD", poresize1, mu1, sigma1)
        plot_psd(ax4, "Better Monopore PSD", poresize2, mu2, sigma2)
        fig.tight_layout()
        ret = plt.show()
    if not ret:
        return


"""
    Models.Stochastic.LognormalUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import root

def compute_mode(mu, sigma):
    return np.exp(mu - sigma**2)

def compute_stdev(mu, sigma):
    return np.sqrt((np.exp(sigma**2) - 1)*np.exp(2*mu + sigma**2))

def compute_mu_sigma(mode, stdev):
    def fun(x):
        mu, var = x
        mode_ = np.exp(mu - var)
        stdev_ = np.sqrt((np.exp(var) - 1)*np.exp(2*mu + var))
        return np.array([mode_ - mode, stdev_ - stdev])

    """
    See https://en.wikipedia.org/wiki/Log-normal_distribution
    We can solve mu, sigma from Mean and Variance,
    while solving from Mode and Variance seems difficult or impossible.
    (sympy cannot give a simple solution)
    """
    M = mode**2     # using mode while M = mean**2 is accurate
    mu_approx = np.log(M/np.sqrt(M + stdev**2))
    V = stdev**2
    var_approx = np.log(1 + V/M)
    res = root(fun, np.array([mu_approx, var_approx]))  # solving variance instead of sigma to avoid negative results.
    mu, var = res.x
    return mu, np.sqrt(var)

def compute_mu_sigma_from_mean(mean, stdev):
    # see https://en.wikipedia.org/wiki/Log-normal_distribution
    # the followins was genarated by Cody AI
    return np.log(mean**2/np.sqrt(mean**2 + stdev**2)), np.sqrt(np.log(1 + stdev**2/mean**2))

def compute_mean_from_mode(mode, stdev):
    mu, sigma = compute_mu_sigma(mode, stdev)
    return np.exp(mu + sigma**2/2)

if __name__ == "__main__":
    mode = compute_mode(6, 0.05)
    stdev = compute_stdev(6, 0.05)
    print("log(mode)=", np.log(mode))
    print("log(stdev)=", np.log(stdev))
    print(mode, stdev, compute_mu_sigma(mode, stdev))
    print(mode, stdev, compute_mean_from_mode(mode, stdev), np.exp(6 + 0.05**2/2))
    mean = compute_mean_from_mode(mode, stdev)
    mu, sigma = compute_mu_sigma_from_mean(mean, stdev)
    print("mean, mu, sigm", mean, mu, sigma)
    for mode in [400, 2.20742076e+02]:
        print("----------------")
        stdev = mode*0.1
        mean = compute_mean_from_mode(mode, stdev)
        print("mode, stdev, mean", mode, stdev, mean)





"""
    UvCorrectorSpline.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from scipy.optimize import curve_fit
from ScipyUtils import get_spline
from molass_legacy.Peaks.ElutionModels import egh
import molass_legacy.KekLib.DebugPlot as plt
from .UvCorrector import UvCorrector

class UvCorrectorSpline(UvCorrector):
    def __init__(self, curve1, curve2):
        self.curve1 = curve1
        self.curve2 = curve2
        self.x = curve2.x
        self.y = curve2.y
        self.d1 = curve1.d1

    def derivative_curve(self, x, scale):
        return scale * self.d1(x)

    def correction_curve(self, x, *params):
        slope, intercept, scale = params
        return slope * x + intercept + self.derivative_curve(x, scale)

    def fit(self, peak_region, debug=False, fig_file=None):
        x = self.x
        y = self.y

        fit_slice = self.get_widest_possible_fit_slice(peak_region)

        fx = x[fit_slice]
        fy = y[fit_slice]
        slope, intercept = linregress(fx, fy)[0:2]

        pos_p0 = np.array([slope, intercept, +1])
        neg_p0 = np.array([slope, intercept, -1])

        popts = []
        errors = []
        for k, p0 in enumerate([pos_p0, neg_p0]):
            if debug:
                from .UvCorrectorIllust import debug_plot
                debug_plot("Initial Parameters", self, x, y, p0, fit_slice, fig_file=fig_file)

            try:
                popt, pcov = curve_fit(self.correction_curve, fx, fy, p0)
                error = np.std(self.correction_curve(fx, *p0) - fy)
                if debug:
                    debug_plot("Optimized Parameters", self, x, y, popt, fit_slice, init_params=p0, fig_file=fig_file)
            except Exception as exc:
                print(exc)
                popt = None
                error = np.inf

            popts.append(popt)
            errors.append(error)

        print("errors=", errors)

        self.popt = popts[0] if errors[0] < errors[1] else popts[1]



"""
    Optimizer.ParamsSelector.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

class ParamsSelector(Dialog):
    def __init__(self, parent, optimizer, params):
        self.optimizer = optimizer
        self.params = params
        self.selection = None
        Dialog.__init__(self, parent, "Parameter Selection", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        label = Tk.Label(body_frame, text="Select exactly two parameters for analysis.", bg="white")
        label.pack(fill=Tk.X, pady=10)
        self.psheet = self.optimizer.params_type.get_params_sheet(body_frame, self.params, None, self.optimizer)
        self.psheet.enable_selection()
        self.psheet.pack()

    def set_selection(self, sel_list):
        self.psheet.set_selection(sel_list)

    def get_selection(self):
        return self.selection

    def ok(self):
        try:
            self.selection = self.psheet.get_selection()    # get it before destruction
        except ValueError:
            return
        Dialog.ok(self)



# coding=utf-8
"""
Example: Radial distribution function g(r)
==========================================

Calculating *g(r)* (radial distribution function) of water, taking
into account periodic boundaries.

Contains a few speed-ups over the most naive implementation
 - use self_distance_array() instead of distance_array() and pre-allocate
   dist array
 - use numpy in-place operations where possible
 - use 1D histogram function (instead of e.g. histogramdd())

Profiling shows that the computational bottleneck is the
:func:`numpy.histogram` function.
"""



import numpy as np

from MDAnalysis import *
from MDAnalysis.lib.distances import *  # #distance_array
import MDAnalysis.units  # for bulk water density

try:
    import matplotlib

    matplotlib.use('agg')  # no interactive plotting, only save figures
    import pylab

    have_matplotlib = True
except ImportError:
    have_matplotlib = False


# very short trajectory for testing; use your own!
# NOTE: not a good example because the trajectory uses a
# dodecahedral periodic box, which MDAnalysis does NOT
#       handle correctly at the moment!!
from MDAnalysis.tests.datafiles import GRO, XTC

universe = Universe(GRO, XTC)

# adjust the selection, e.g. "resname TIP3 and name OH2" for CHARMM
solvent = universe.select_atoms("resname SOL and name OW")

dmin, dmax = 0.0, 8.0
nbins = 80

# set up rdf
rdf, edges = np.histogram([0], bins=nbins, range=(dmin, dmax))
rdf *= 0
rdf = rdf.astype(np.float64)  # avoid possible problems with '/' later on

print(dir(solvent))

# n = solvent.numberOfAtoms()
n = len(solvent.atoms)
dist = np.zeros((n * (n - 1) // 2,), dtype=np.float64)

print("Start: n = {0:d}, size of dist = {1:d} ".format(n, len(dist)))

boxvolume = 0
for ts in universe.trajectory:
    print("Frame {0:4d}".format(ts.frame))
    boxvolume += ts.volume  # correct unitcell volume
    coor = solvent.coordinates()
    # periodicity is NOT handled correctly in this example because
    # distance_array() only handles orthorhombic boxes correctly
    ##box = ts.dimensions[:3]     # fudge: only orthorhombic boxes handled correctly
    # DISABLE:
    box = None
    self_distance_array(coor, box, result=dist)  # use pre-allocated array, box not fully correct!!
    new_rdf, edges = np.histogram(dist, bins=nbins, range=(dmin, dmax))
    rdf += new_rdf
print()

numframes = universe.trajectory.numframes / universe.trajectory.skip
boxvolume /= numframes  # average volume

# Normalize RDF
radii = 0.5 * (edges[1:] + edges[:-1])
vol = (4. / 3.) * np.pi * (np.power(edges[1:], 3) - np.power(edges[:-1], 3))
# normalization to the average density n/boxvolume in the simulation
density = n / boxvolume
# This is inaccurate when solutes take up substantial amount
# of space. In this case you might want to use
## import MDAnalysis.units
## density = MDAnalysis.units.convert(1.0, 'water', 'Angstrom^{-3}')
norm = density * (n - 1) / 2 * numframes
rdf /= norm * vol

outfile = './output/rdf.dat'
with open(outfile, 'w') as output:
    for radius, gofr in zip(radii, rdf):
        output.write("{radius:8.3f} \t {gofr:8.3f}\n".format(**vars()))
print("g(r) data written to {outfile!r}".format(**vars()))

if have_matplotlib:
    matplotlib.rc('font', size=14)
    matplotlib.rc('figure', figsize=(5, 4))
    pylab.clf()
    pylab.plot(radii, rdf, linewidth=3)
    pylab.xlabel(r"distance $r$ in $\AA$")
    pylab.ylabel(r"radial distribution function $g(r)$")
    pylab.savefig("./figures/rdf.pdf")
    pylab.savefig("./figures/rdf.png")
    print("Figure written to ./figures/rdf.{pdf,png}")



"""
    Theory.RotationMatrix.py

    from
        Random Rotation Matrix in Python
        http://blog.lostinmyterminal.com/python/2015/05/12/random-rotation-matrix.html
"""
import numpy as np

def rand_rotation_matrix(deflection=1.0, randnums=None):
    """
    Creates a random rotation matrix.
    
    deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
    rotation. Small deflection => small perturbation.
    randnums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
    """
    # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c
    
    if randnums is None:
        randnums = np.random.uniform(size=(3,))
        
    theta, phi, z = randnums
    
    theta = theta * 2.0*deflection*np.pi  # Rotation about the pole (Z).
    phi = phi * 2.0*np.pi  # For direction of pole deflection.
    z = z * 2.0*deflection  # For magnitude of pole deflection.
    
    # Compute a vector V used for distributing points over the sphere
    # via the reflection I - V Transpose(V).  This formulation of V
    # will guarantee that if x[1] and x[2] are uniformly distributed,
    # the reflected points will be uniform on the sphere.  Note that V
    # has length sqrt(2) to eliminate the 2 in the Householder matrix.
    
    r = np.sqrt(z)
    Vx, Vy, Vz = V = (
        np.sin(phi) * r,
        np.cos(phi) * r,
        np.sqrt(2.0 - z)
        )
    
    st = np.sin(theta)
    ct = np.cos(theta)
    
    R = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))
    
    # Construct the rotation matrix  ( V Transpose(V) - I ) R.
    
    M = (np.outer(V, V) - np.eye(3)).dot(R)
    return M

def rotate_3d_mesh(M, center, x, y, z):
    assert x.shape == y.shape and y.shape == z.shape
    x_ = x - center[0]
    y_ = y - center[1]
    z_ = z - center[2]
    P = np.array([x_, y_, z_]).reshape(3, x.shape[0]*x.shape[1])
    R = M @ P
    return (R + np.array(center)[:,np.newaxis]).reshape((3, *x.shape))



"""
    RgProcess.RgCurveProxy.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import os
import logging
from time import time
import numpy as np  # used in the eval, in cases where the value is like 'np.int64(140)'
from numpy import array, int64      # array and int64 are referenced during the eval below
from molass_legacy.RgProcess.RgCurve import RgCurve
from molass_legacy.Trimming import TrimmingInfo   # used in eval_file

class RgCurveProxy(RgCurve):
    def __init__(self, ecurve, rg_folder, progress_cb=None):
        self.logger = logging.getLogger(__name__)

        t0 = time()

        if ecurve is not None:
            self.x = ecurve.x
            self.y = ecurve.y
        self.ecurve = ecurve
        self.X = None
        self.excl_info = None
        self.excl_spline = None

        objects = []
        for filename in ["slices.txt", "states.txt", "segments.txt", "qualities.txt", "rg_trimming.txt", "baseline_type.txt"]:
            path  = os.path.join(rg_folder, filename)
            if os.path.exists(path):
                from molass_legacy.KekLib.EvalUtils import eval_file
                objects.append(eval_file(path, locals_=globals()))
            else:
                objects.append(None)    # some files may not exist

        self.slices = objects[0]
        self.states = objects[1]
        self.segments = objects[2]
        self.qualities = objects[3] if len(objects) > 3 else None
        self.rg_trimming = objects[4] if len(objects) > 4 else None
        self.baseline_type = objects[5] if len(objects) > 5 else None

        self.logger.info("It took %.3g seconds for rg_curve proxy construction.", time()-t0)



# coding: utf-8
"""
    Optimizer.EghGradient.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Models.ElutionCurveModels import egh

PARAMS = ['h', 'mu', 'sigma', 'tau']    # 'a' not used
NUM_PARAMS = len(PARAMS)

def compute_gradient_impl(M, P, rank, x, pv):
    print('compute_gradient_impl')
    pv_ = pv.reshape((rank, NUM_PARAMS))
    DP = np.zeros((NUM_PARAMS, len(x)))
    C_list = []
    for h, mu, sigma, tau in pv_:
        fv = egh(x, h, mu, sigma, tau)
        dh = egh(x, 1, mu, sigma, tau)
        x_mu = x-mu
        denom1 = 2*sigma**2 + tau*x_mu
        denom2 = denom1**2
        dm = fv * (-tau*x_mu**2/denom2 + 2*x_mu/denom1)
        ds = fv * 4*sigma*x_mu**2/denom2
        dt = fv * x_mu**3/denom2
        C_list.append(fv)
        DP += np.array([dh, dm, ds, dt])    # ok?
    C = np.array(C_list)
    DC = 2 * P.T @ (P@C - M)
    print('DC.shape=', DC.shape)
    print('DP.shape=', DP.shape)
    ret = (DC @ DP.T).flatten()
    print('ret.shape=', ret.shape)
    return ret



# coding: utf-8
"""
    LowRankSolver.py

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
import copy


class LowRankSolver:
    def __init__( self, data_matrix ):
        self.M  = data_matrix

    def solve_base_plane( self, cols ):
        rank = len(cols)

        BP = self.generate_baseplane( A, B, C )
        M_ = self.M - BP
        P = col_curve_matrix - BP[:,cols]
        Pinv = np.linalg.pinv( P )
        C = np.dot( Pinv, M_ )

    def generate_baseplane( self, A, B, C ):
        pass



"""
    QuickAnalysis.ModeledPeaksTester.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy._MOLASS.SerialSettings import get_setting

class TestDialog(Dialog):
    def __init__(self, parent, sd):
        self.sd = sd
        Dialog.__init__(self, parent, "", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        sbox_frame = Tk.Frame(body_frame)
        sbox_frame.pack()

        label = Tk.Label(sbox_frame, text="Num Peaks")
        label.pack(side=Tk.LEFT)

        xr_curve = self.sd.get_xray_curve()
        n = len(xr_curve.peak_info)

        self.num_peaks = Tk.IntVar()
        self.num_peaks.set(n+1)
        sbox  = Tk.Spinbox(sbox_frame, textvariable=self.num_peaks,
                          from_=1, to=7, increment=1,
                          justify=Tk.CENTER, width=6)
        sbox.pack(side=Tk.LEFT)

        self.debug = Tk.IntVar()
        debug_cb = Tk.Checkbutton(body_frame, text="debug", variable=self.debug)
        debug_cb.pack()
        run_btn = Tk.Button(body_frame, text="run", command=self.run)
        run_btn.pack()

    def run(self):
        from importlib import reload
        import QuickAnalysis.ModeledPeaks
        reload(QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import get_curve_xy_impl, get_modeled_peaks_impl

        unified_baseline_type = get_setting("unified_baseline_type")
        print("unified_baseline_type=", unified_baseline_type)
        uv_x, uv_y, xr_x, xr_y, details = get_curve_xy_impl(self.sd, baseline_type=unified_baseline_type, return_details=True)
        num_peaks = self.num_peaks.get()
        debug = self.debug.get()
        a, b = self.sd.pre_recog.mapped_info[0]
        uv_peaks, xr_peaks = get_modeled_peaks_impl(a, b, uv_x, uv_y, xr_x, xr_y, num_peaks=num_peaks, exact_num_peaks=num_peaks, debug=debug)
        plot_modeled_peaks(uv_x, uv_y, xr_x, xr_y, uv_peaks, xr_peaks, a, b, details=details)

def plot_modeled_peaks(uv_x, uv_y, xr_x, xr_y, uv_peaks, xr_peaks, a, b, details=None, suptitle="unspecified title"):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.Peaks.ElutionModels import egh

    with plt.Dp():
        fig, axes = plt.subplots(ncols=3, figsize=(18,5))
        fig.suptitle(suptitle)
        ax1, ax2, ax3 = axes
        ax1.set_title("UV Elution")
        ax2.set_title("Xray Elution")
        ax3.set_title("UV/Xray Mapping")

        ax1.plot(uv_x, uv_y, color="blue")
        uv_ty = np.zeros(len(uv_y))
        for h, m, s, t in uv_peaks:
            uv_cy = egh(uv_x, h, m, s, t)
            uv_ty += uv_cy
            ax1.plot(uv_x, uv_cy, ":")
            ax1.plot(m, h, "o", color="yellow")
        ax1.plot(uv_x, uv_ty, ":", color="red")

        ax2.plot(xr_x, xr_y, color="orange")
        xr_ty = np.zeros(len(xr_y))
        for h, m, s, t in xr_peaks:
            xr_cy = egh(xr_x, h, m, s, t)
            xr_ty += xr_cy
            ax2.plot(xr_x, xr_cy, ":")
            ax2.plot(m, h, "o", color="yellow")
        ax2.plot(xr_x, xr_ty, ":", color="red")

        ax3.plot(uv_x, uv_y, color="blue")
        axt = ax3.twinx()
        axt.grid(False)
        axt.plot(xr_x*a + b, xr_y, color="orange")

        if details is not None:
            uv_by, xr_by = details.baselines
            ax1.plot(uv_x, uv_by, color="red")
            ax2.plot(xr_x, xr_by, color="red")

        fig.tight_layout()
        plt.show()



"""
    Optimizer.FvSynthesisOptimizer.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from .FvSynthesizer import synthesize
from .FvScoreConverter import convert_score

# (XR_2D_fitting, XR_LRF_residual, UV_2D_fitting, UV_LRF_residual, Guinier_deviation, Kratky_smoothness, SEC_conformance), sv0, sv1,
VALUE_TABLE = [
        # [(-1.45392, -1.19308,  -1.35939,  -0.076241,  -1.1585, -1.03118,  -0.6), 70.3, 90.],        # from 20230303/HasA
        # [(-1.22972, -0.852194, -0.562419, -0.895275,  -1.1217, -0.992915, -0.6), 71.5, 50.],        # from 20230303/HasA
        [(-0.908609, -1.10815, -0.999691, -0.0794044, -1.12317, -1.07021, -0.6), 71.1, 90],         # from 20230303/HasA
        [(-0.894142, -0.996312, -0.37274, -0.78015,   -1.11888, -1.04737,  -0.6), 71.4, 50.],       # from 20230303/HasA
    ]

def objective_func(w):
    w1 = 1 - np.sum(w)
    w_ = np.concatenate([w, [w1]])
    svdev = 0
    for v, sv0, sv1 in VALUE_TABLE:
        fv = synthesize(v, positive_elevate=3, value_weights=w_)
        sv = convert_score(fv)
        svdev += (sv - sv1)**2 + (sv - sv0)**2
    return np.log10(svdev) + max(-1.5, np.log10(np.std(w_)))
    # return np.log10(svdev)

def compute_optimized_weights():
    from molass_legacy._MOLASS.SerialSettings import get_setting

    NUM_MAJOR_SCORES = get_setting("NUM_MAJOR_SCORES")
    init_weights = np.ones(NUM_MAJOR_SCORES)/NUM_MAJOR_SCORES

    bounds = [(0, 1)] * (NUM_MAJOR_SCORES - 1)
    ret = minimize(objective_func, init_weights[0:-1], bounds=bounds)

    weights = np.concatenate([ret.x, [1 - np.sum(ret.x)]])
    # [0.13661366 0.13661366 0.15548285 0.13661366 0.13661366 0.13661366 0.16144883]

    # print(ret.fun)
    show_weights_performance(weights)

def show_weights_performance(weights):
    print(weights)
    for k, (v, sv0, sv1) in enumerate(VALUE_TABLE):
        fv = synthesize(v, positive_elevate=3)
        sv = float("%.3g" % convert_score(fv))
        print([k], sv0, sv1, sv)



"""
    Asymptotical.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt

def asymptotical(x, y, dinit, rate, return_d=False):
    """
    x : log value like guinier deviation
    y : log value like LRF residual

    d : should decrease as x and y impove with smaller dist(x, y)
    (x - (y + d)) : should be evaluated smaller as they improve
    """
    d = np.exp((y + dinit)*rate)
    z = d*(x - (y + d))**2

    if return_d:
        return z, d
    else:
        return z

if __name__ == '__main__':
    from matplotlib import cm

    x = np.linspace(-3, 1, 100)
    y = np.linspace(-3, 1, 100)
    # z = asymptotical(x, y, 0.5, 0.1)
    xx, yy = np.meshgrid(x, y)
    zz, dd = asymptotical(xx, yy, 0.5, 1, return_d=True)

    fig = plt.figure(figsize=(12,5))
    ax1 = fig.add_subplot(121, projection="3d")
    ax2 = fig.add_subplot(122)

    ax1.set_title("(x, y) Deviation Evaluation")
    ax1.set_xlabel("x")
    ax1.set_ylabel("y")
    ax1.set_zlabel("z")
    ax1.plot_surface(xx, yy, zz, cmap=cm.coolwarm)

    ax2.set_title("Asymptotic Differance")
    ax2.set_xlabel("y")
    ax2.set_ylabel("d")

    ax2.plot(y, dd)

    fig.tight_layout()
    plt.show()



"""
    Affine.py

    Copyright (c) 2018-2023, Masatsuyo Takahashi, KEK-PF
"""

import numpy                as np

NUM_POINTS  = 3

class Affine:
    def __init__( self, src_points, tgt_points, raise_=False):
        """
        M   [
                [ a, b ]
                [ c, d ]
            ]

        v   [ e, f ]

        p -> q ; q = np.dot( M, p ) + v

        a * p[0] + b * p[1] + c *  0   + d *  0   + e * 1 + f * 0 = q[0]
        a *  0   + b *  0   + c * p[0] + d * p[1] + e * 0 + f * 1 = q[1]

        """
        assert len(src_points) == NUM_POINTS
        assert len(tgt_points) == NUM_POINTS
        A_list  = []
        b_list  = []
        for i, p in enumerate( src_points ):
            p0, p1  = p
            q0, q1  = tgt_points[i]
            #                 a   b   c   d   e   f
            A_list.append( [ p0, p1,  0,  0,  1,  0 ] )
            b_list.append( q0 )
            A_list.append( [  0,  0, p0, p1,  0,  1 ] )
            b_list.append( q1 )

        A   = np.array( A_list )
        b   = np.array( b_list )

        try:
            x   = np.linalg.solve( A, b )
        except Exception as exc:
            import logging
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            logger = logging.getLogger(__name__)
            if raise_:
                msg = "Affine Error: "
            else:
                msg = "later affine.transform(x, y) will return the same x, y for singular cases: "
            log_exception(logger, msg)
            logger.info('src_points=' + str(src_points))
            logger.info('tgt_points=' + str(tgt_points))

            if raise_:
                raise( exc )
            else:
                self.is_singlar = True
                return

        self.is_singlar = False
        self.matrix = np.array( [ [ x[0], x[1] ], [ x[2], x[3] ] ] )
        self.vector = np.array( [ x[4], x[5] ] )
        self.matrix_1 = np.array( [ [ x[0], x[1], x[4] ], [ x[2], x[3], x[5] ] ] )

    def transform_list( self, points ):
        return [ np.dot( self.matrix, point ) + self.vector for point in points ]

    def transform( self, x, y ):
        if self.is_singlar:
            return x, y

        tp = np.dot( self.matrix_1, np.array( [ x, y, np.ones(len(x)) ] ) )
        return tp[0,:], tp[1,:]



"""
    Batch.FullBatch.py

    eventually to be the baseclass of V2 procedures

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import re
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, clear_temporary_settings
from molass_legacy.Peaks.PeakParamsSet import PeakParamsSet

class FullBatch:
    def __init__(self):
        # to be moved from molass_legacy.Peaks.PeakEditor
        self.unified_baseline_type = get_setting("unified_baseline_type")
        self.elution_model = get_setting("elution_model")
        self.ecurve_info = None

    def load_data(self, in_folder):
        from molass_legacy.Batch.StandardProcedure import StandardProcedure
        self.logger.info("loading data from %s", in_folder)
        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)
        return sd

    def prepare(self, sd, num_peaks=None, min_num_peaks=None, debug=False):
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        from molass_legacy.Optimizer.DefaultNumPeaks import get_default_num_peaks
        from SecSaxs.DataTreatment import DataTreatment
        self.pre_recog = PreliminaryRecognition(sd)
        if num_peaks is None:
            num_peaks = get_default_num_peaks(sd)
            if min_num_peaks is not None:
                num_peaks = max(min_num_peaks, num_peaks)

        self.logger.info("preparing as num_peaks=%d", num_peaks)

        treat = DataTreatment(route="v2", correction=1)
        trimmed_sd = treat.get_trimmed_sd(sd, self.pre_recog, debug=debug)

        # set PeakEditor attributes
        self.sd = trimmed_sd
        self.corrected_sd = treat.get_corrected_sd(sd, self.pre_recog, trimmed_sd)
        self.base_curve_info = treat.get_base_curve_info()
        self.exact_num_peaks = num_peaks
        self.ecurve_info = None
        self.fullopt_class = None

        """
        task: remove this (may be irrelevant)
        uv_x, uv_y, xr_x, xr_y, baselines = self.get_curve_xy(return_baselines=True)
        uv_peaks, xr_peaks = self.get_modeled_peaks(uv_x, uv_y, xr_x, xr_y)
        """

    def get_pre_recog_mapping_params(self):
        cs = self.pre_recog.cs
        return cs.slope, cs.intercept

    def get_curve_xy(self, return_baselines=False, debug=False):
        if debug:
            import QuickAnalysis.ModeledPeaks
            from importlib import reload
            reload(QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import get_curve_xy_impl

        if self.ecurve_info is None:
            self.ecurve_info = get_curve_xy_impl(self.sd, baseline_type=self.unified_baseline_type, return_details=True, debug=debug)
            details = self.ecurve_info[-1]
            self.ecurves = details.uv_curve, details.xr_curve
            self.baseline_objects = details.baseline_objects
            self.baselines = details.baselines
            # self.baselines[0] = self.get_uv_baseline()    # get_uv_baseline_deprecated
            self.baseline_params = details.baseline_params
            self.logger.info("got xr baseline params as %s with unified_baseline_type=%d", str(self.baseline_params[1]), self.unified_baseline_type)

        if return_baselines:
            return *self.ecurve_info[0:4], self.baselines
        else:
            return self.ecurve_info[0:4]

    def get_uv_baseline(self, xy=None):
        # task: verify that this is correct
        if xy is None:
            return self.baselines[0]
        else:
            raise NotImplementedError("get_uv_baseline with xy argument is not implemented yet. Use get_uv_baseline_deprecated instead.")

    def get_uv_baseline_deprecated(self, xy=None):
        if xy is None:
            uv_x, uv_y = self.ecurve_info[0:2]
        else:
            uv_x, uv_y = xy
        uv_base_curve, uv_baseparams = self.base_curve_info
        intreg_scale, y_ = uv_base_curve.guess_integral_scale(uv_x, uv_y, uv_baseparams[0:7])
        uv_baseparams[-1] = intreg_scale    # note that this updated params will be used later in self.get_uv_base_params()
        return uv_base_curve(uv_x, uv_baseparams, y_)

    def get_modeled_peaks(self, uv_x, uv_y, xr_x, xr_y, num_peaks=None, affine=True, min_area_prop=None, debug=False):
        if debug:
            import molass_legacy.QuickAnalysis.ModeledPeaks
            from importlib import reload
            reload(molass_legacy.QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import get_modeled_peaks_impl
        if num_peaks is None:
            num_peaks = self.exact_num_peaks
        a, b = self.get_pre_recog_mapping_params()
        uv_peaks, xr_peaks = get_modeled_peaks_impl(a, b, uv_x, uv_y, xr_x, xr_y, num_peaks, exact_num_peaks=num_peaks,
                                                affine=affine, min_area_prop=min_area_prop, debug=debug)        
        self.peak_params_set = PeakParamsSet(uv_peaks, xr_peaks, a, b)     # for backward compatibility
        return uv_peaks, xr_peaks

    def get_peak_params_set(self):
        return self.peak_params_set

    def set_lrf_src_args1(self, uv_x, uv_y, xr_x, xr_y, baselines):
        self.lrf_src_args1 = (uv_x, uv_y, xr_x, xr_y, baselines)

    def get_lrf_source(self, in_folder=None, min_num_peaks=None, clear=True, use_mapping=False, devel=False):
        if in_folder is None:
            """
            if in_folder is None,
            it is currently assumed that the lrf_src_args1 and peak_params_set are set
            as in Peaks.PeakEditor or Batch.DataBridge
            """
        else:
            """
            it is currently assumed that this is called as a LiteBatch method
            """
            if clear:
                clear_temporary_settings()      # this is usually harmless, be careful though.
            sd = self.load_data(in_folder)
            self.prepare(sd, min_num_peaks=min_num_peaks)
            uv_x, uv_y, xr_x, xr_y, baselines = self.get_curve_xy(return_baselines=True)
            uv_y_ = uv_y - baselines[0]
            xr_y_ = xr_y - baselines[1]
            if use_mapping:
                # note that this is a method of the MappingBatch class
                self.modelled_peaks_method(uv_x, uv_y_, xr_x, xr_y_)
            else:
                uv_peaks, xr_peaks = self.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_, affine=True, min_area_prop=None)               
            self.set_lrf_src_args1(uv_x, uv_y, xr_x, xr_y, baselines)  # task: do not require users to do this
            set_setting('in_folder', in_folder)

        if devel:
            from importlib import reload
            import molass_legacy.Selective.LrfSource
            reload(molass_legacy.Selective.LrfSource)
        from molass_legacy.Selective.LrfSource import LrfSource
        return LrfSource(self.sd, self.corrected_sd, self.lrf_src_args1, self.peak_params_set, pre_recog=self.pre_recog)

    def get_n_components(self):
        num_base_elements = 1
        return len(self.peak_params_set[1]) + num_base_elements

    def get_function_class(self):
        if self.fullopt_class is None:
            fullopt_key = self.key_list[self.func_info.default_index]
            fullopt_class = self.func_dict[fullopt_key]
            code_re = re.compile(r"^(\w+)\s*:")
            m = code_re.search(fullopt_key)
            if m:
                class_code = m.group(1)
            else:
                assert False
        else:
            fullopt_class = self.fullopt_class
            class_code = self.class_code
        return fullopt_class, class_code

    def get_model_name(self):
        from molass_legacy.Optimizer.OptimizerUtils import get_model_name
        fullopt_class, class_code = self.get_function_class()
        return get_model_name(class_code)

    def construct_optimizer(self, fullopt_class=None):
 
        if fullopt_class is None:
            fullopt_class, class_code = self.get_function_class()
        n_components = self.get_n_components()
        """
        if self.base_curve_info is None:
            uv_base_curve = None
        else:
            uv_base_curve = self.base_curve_info[0]
        """
        uv_base_curve = self.baseline_objects[0]
        xr_base_curve = self.baseline_objects[1]
        self.uv_base_curve = uv_base_curve
        
        self.optimizer = fullopt_class(
            self.dsets,
            n_components,
            uv_base_curve=uv_base_curve,
            xr_base_curve=xr_base_curve,
            qvector=self.sd.qvector,    # trimmed sd
            wvector=self.sd.lvector,
            )

        self.fullopt = self.optimizer   # for backward compatibility
        self.params_type = self.fullopt.params_type

    def get_optimizer(self):
        return self.optimizer

    def compute_init_params(self, debug=False):

        estimator = self.params_type.get_estimator(self)

        try:
            self.init_params = estimator.estimate_params(debug=debug)
        except:
            # for classes not supporing debug
            self.init_params = estimator.estimate_params()

        self.logger.info("len(init_params)=%d, init_params[-7:]=%s", len(self.init_params), str(self.init_params[-7:]))

        return self.init_params

    def get_mapped_params(self):
        uv_peaks, xr_peaks, a, b = self.peak_params_set
        num_peaks_mapping = len(xr_peaks)
        return a, b, num_peaks_mapping

    def get_uv_base_params(self, xyt=None, debug=False):
        if debug:
            import Baseline.UvBaseline
            from importlib import reload
            reload(Baseline.UvBaseline)
        from molass_legacy.Baseline.UvBaseline import get_uv_base_params_impl
        return get_uv_base_params_impl(self, xyt=xyt, debug=debug)

if __name__ == '__main__':
    pass



# coding: utf-8
"""
    DataAverager.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
import difflib
from SerialDataUtils import load_xray_files
from molass_legacy.KekLib.NumpyUtils import np_savetxt
from InputSmootherAveraging import IntensitySmootherAveraging
from molass_legacy._MOLASS.SerialSettings import set_setting, get_setting
import molass_legacy.KekLib.CustomMessageBox as MessageBox

def average_impl(num_average, in_files1, out_folder, name_pattern, dialog=None, debug=False):
    num_progress = 0
    num_load_files = 50
    num_average_files = 10

    if dialog is not None:
        dialog.mpb["maximum"] = num_load_files + num_average_files + len(in_files1)
        dialog.mpb["value"] = num_progress

    set_setting('found_lacking_q_values', 0)
    data_array1, comments1, lacking_info1 = load_xray_files(in_files1, return_lacking_info=True)
    if get_setting('found_lacking_q_values'):
        MessageBox.showwarning( "Lacking Q-values",
            "There have been found lacking Q-values\n"
            "in some of the input files.\n"
            "Make sure to confirm them in the molass.log",
            parent=dialog )

    if dialog is not None:
        num_progress += num_load_files
        dialog.mpb["value"] = num_progress
        dialog.update()

    smoother = IntensitySmootherAveraging(data_array1, num_average)
    indeces = np.arange( 0, data_array1.shape[0] )     # these indeces start at 0, not at self.xr_j0
    averaged_array, average_slice_array = smoother( indeces, return_numpy_ndarray=True )

    if dialog is not None:
        num_progress += num_average_files
        dialog.mpb["value"] = num_progress
        dialog.update()

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        from MatrixData import simple_plot_3d
        plt.push()
        fig = plt.figure(figsize=(14, 7))
        ax1 = fig.add_subplot(121, projection='3d')
        ax2 = fig.add_subplot(122, projection='3d')
        simple_plot_3d(ax1, data_array1[:,:,1].T)
        simple_plot_3d(ax2, averaged_array[:,:,1].T)
        plt.show()
        plt.pop()

    out_data_list = []
    for i, data in enumerate( averaged_array ):
        filename = name_pattern % i
        # print( 'saving', filename )
        filepath = os.path.join(out_folder, filename)
        with open( filepath, "wb" ) as fh:
            slice_ = average_slice_array[i]
            fh.write( str.encode( '# Created by averaging the following %d files with []-numbering starting from 1.\n' % ( slice_.stop - slice_.start ) ) )
            for j in range( slice_.start, slice_.stop ):
                fh.write( str.encode( '# [%d] %s\n' % ( j+1, in_files1[j] ) ) )
            fh.write( str.encode( "#\n# Q\tIntensity\tError\n" ) )
            fh.close()
        np_savetxt( filepath, data, mode="a" )
        out_data_list.append(data)
        if dialog is not None:
            n = i+1
            dialog.num_files3.set(n)
            dialog.mpb["value"] = num_progress + n
            dialog.update()

    return out_data_list



"""
    PeakMappingSolver.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from sklearn.cluster import KMeans
from molass_legacy.Elution.CurveUtils import simple_plot

PENALTY_DIFF_RATIO = 0.1    # > 0.05 which is too small for 20161104/BL-6A/HIF

class PeakMappingSolver:
    def __init__(self, a_curve, x_curve, mapping, debug=False):
        self.logger = logging.getLogger(__name__)

        uv_x = a_curve.x
        uv_y = a_curve.y

        self.penalty_diff = len(uv_x)* PENALTY_DIFF_RATIO

        A, B = mapping
        xr_x = x_curve.x
        xr_y = x_curve.y
        x = A*xr_x + B

        uv_pt_x = a_curve.peak_top_x
        xr_pt_x = x_curve.peak_top_x

        mp_pt_x = A*xr_pt_x + B
        points = np.concatenate([uv_pt_x, mp_pt_x])
        X = points.reshape((len(points), 1))
        if debug:
            print("points=", points)
            scale = a_curve.max_y/x_curve.max_y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("PeakMappingSolver (1)")
                simple_plot(ax, a_curve)
                ax.plot(x, scale*xr_y, ":")
                fig.tight_layout()
                plt.show()

        uv_np = len(uv_pt_x)
        xr_np = len(xr_pt_x)
        max_n = max(uv_np, xr_np)

        min_score = None
        opt_mappers = None
        for n in range(max_n, max_n+2):
            kmeans = KMeans(n_clusters=n, random_state=0).fit(X)
            labels = kmeans.labels_
            score, mappers = self.evaluate_grouping(n, uv_pt_x, mp_pt_x, labels[:uv_np], labels[uv_np:], debug=debug)
            if debug:
                print([n], "labels=", labels, "score=", score, "mappers=", mappers)
            if min_score is None or score < min_score:
                min_score = score
                opt_mappers = mappers

        if debug:
            print("min_score=", min_score)
            print("opt_mappers=", opt_mappers)
            uv_mapper, xr_mapper = opt_mappers
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(20,5))
                fig.suptitle("PeakMappingSolver (2)")
                simple_plot(ax1, a_curve, color="blue", legend=False)
                ax1.plot(x, xr_y, ":", label="mapped")
                for px in x_curve.peak_top_x:
                    mx = A*px + B
                    ax1.plot(mx, x_curve.spline(px), "o", color="yellow")
                ax1.legend()
                for i, j in enumerate(uv_mapper):
                    if j < 0:
                        continue
                    px = uv_pt_x[i]
                    ax1.plot(px, a_curve.spline(px), "o", color="cyan", markersize=20, alpha=0.2)

                simple_plot(ax2, x_curve, color="orange")

                for i, j in enumerate(xr_mapper):
                    if j < 0:
                        continue
                    px = xr_pt_x[i]
                    ax2.plot(px, x_curve.spline(px), "o", color="cyan", markersize=20, alpha=0.2)

                fig.tight_layout()
                plt.show()

        self.opt_mappers = opt_mappers

    def get_opt_mappers(self):
        return self.opt_mappers

    def evaluate_grouping(self, n, uv_pt_x, mp_pt_x, uv_labels, xr_labels, debug=False):

        uv_dict = dict([ (j, i) for i, j in enumerate(uv_labels) ])
        xr_dict = dict([ (j, i) for i, j in enumerate(xr_labels) ])

        if debug:
            print([n], "uv_labels=", uv_labels)
            print([n], "xr_labels=", xr_labels)
            print("uv_dict=", uv_dict)
            print("xr_dict=", xr_dict)

        uv_mapper = np.ones(len(uv_pt_x))*(-1)
        xr_mapper= np.ones(len(mp_pt_x))*(-1)

        total_diff = 0
        match_count = 0
        for label in range(n):
            uv_i = uv_dict.get(label)
            xr_i = xr_dict.get(label)
            if uv_i is None and xr_i is None:
                # assert False                  # true for 20191031
                self.logger.info("unexpected label %d in uv_labels=%s", label, uv_labels)
                self.logger.info("unexpected label %d in xr_labels=%s", label, xr_labels)
                diff = self.penalty_diff
            elif uv_i is not None and xr_i is not None:
                uv_mapper[uv_i] = xr_i
                xr_mapper[xr_i] = uv_i
                diff = (uv_pt_x[uv_i] - mp_pt_x[xr_i])**2
                match_count += 1
            else:
                diff = self.penalty_diff
            if debug:
                print([label], "diff=", diff)
            total_diff += diff

        score = (n - match_count)/n * total_diff

        if debug:
            print("---- uv_mapper=", uv_mapper)
            print("---- xr_mapper=", xr_mapper)
            print("---- match_count=", match_count)
            print("---- total_diff=", total_diff)
            print("---- score=", score)

        return score, (uv_mapper, xr_mapper)



"""
    LinearBaseline.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

USE_END_PARAMS = False
if not USE_END_PARAMS:
    from .Constants import SLOPE_SCALE

class LinearBaseline:
    def __init__(self, x=None, y=None, debug=False):
        if debug:
            from importlib import reload
            import Baseline.Baseline
            reload(Baseline.Baseline)
        from .Baseline import compute_baseline

        if x is None:
            return

        yb, (a,b) = compute_baseline(y, x=x, return_params=True, debug=debug)
        self.yb = yb
        self.params = [a*SLOPE_SCALE, b]
        self.x1 = x[0]
        self.x2 = x[-1]
        self.end_params = yb[[0,-1]]

    def __call__(self, x, params, y_, cy_list):
        # y_ is not used in LinearBaseline
        # cy_list is not used in LinearBaseline

        if USE_END_PARAMS:
            """
            (y2 - y1) = k*(x2 - x1)
            k = (y2 - y1)/(x2 - x1)

            (y - y1) = k*(x - x1)
            y = y1 + k*(x - x1)
            """
            y1, y2 = params
            k = (y2 - y1)/(self.x2 - self.x1)
            return y1 + k*(x - self.x1)
        else:
            slope, intercept = params
            return x*slope/SLOPE_SCALE + intercept



"""
    LogChecker.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import sys
import glob
import re

started_log_re = re.compile(r"started loading (\S+)")
yyyymmdd_re = re.compile(r"^(\d{8}).*")

def check_logfiles(log_root):
    num_files = 0
    num_normals = 0
    num_errors = 0
    first_yyyymmdd = None
    last_yyyymmdd = None
    for file in glob.glob(log_root + "/*/*.log"):
        num_files += 1

        # print(file)
        found_started = False
        found_finished = False
        with open(file) as fh:
            for n, line in enumerate(fh):
                if not found_started:
                    m = started_log_re.search(line)
                    if m:
                        for node in m.group(1).split("/"):
                            m_ = yyyymmdd_re.match(node)
                            if m_:
                                yyyymmdd = m_.group(1)
                                if first_yyyymmdd is None:
                                    first_yyyymmdd = yyyymmdd
                                last_yyyymmdd = yyyymmdd

                if line.find("Finished.") > 0:
                    found_finished = True

        if found_finished:
            print(file, "ok")
            num_normals += 1
        else:
            print(file, "something wrong")
            num_errors += 1

    print("num_files     %4d     %s ～ %s" % (num_files, first_yyyymmdd, last_yyyymmdd))
    print("num_normals   %4d" % num_normals)
    print("num_errors    %4d" % num_errors)

if __name__ == '__main__':
    assert len(sys.argv) > 1
    check_logfiles(sys.argv[1])



# -*- coding: utf-8 -*-
"""

    ZoomPan.py

    Borrowed from
    stack overflow
        "Matplotlib plot zooming with scroll wheel"
        http://stackoverflow.com/questions/11551049/matplotlib-plot-zooming-with-scroll-wheel

    Modified so as to handle multiple subplot objects

"""
from __future__ import print_function
import time
from ControlKeyState    import get_shift_key_state, get_ctrl_key_state
from OurColorMaps       import CmapAlbulaLikeDynamic, Diverging

class ZoomPan:
    def __init__( self, obj_id, canvas, ax, im_rec ):
        self.obj_id = obj_id
        self.canvas = canvas
        self.orig_xlim = ax.get_xlim()
        self.orig_ylim = ax.get_ylim()
        self.press = None
        self.cur_xlim = None
        self.cur_ylim = None
        self.x0 = None
        self.y0 = None
        self.x1 = None
        self.y1 = None
        self.xpress = None
        self.ypress = None
        self.accumulated_scale = 1.0
        self.no_effect_count = 0
        self.im_rec = im_rec
        self.accumulated_cmap_adjustment = 0.0

    def zoom_factory(self, ax_array, base_scale = 2.):
        ax = ax_array[0]

        def cmap_adjust( event ):
            # print( 'obj_id=%d, time=%g' % (self.obj_id, time.time()) )

            if event.button == 'up':
                if self.accumulated_cmap_adjustment < 0.25:
                    self.accumulated_cmap_adjustment += 0.05
                else:
                    return
            else:
                if self.accumulated_cmap_adjustment > -0.25:
                    self.accumulated_cmap_adjustment -= 0.05
                else:
                    return

            im, cmap = self.im_rec
            new_cmap_ = cmap.adjusted_cmap( self.accumulated_cmap_adjustment )
            im.set_cmap( new_cmap_ )
            ax.figure.canvas.draw()

        def zoom(event):
            if event.inaxes not in ax_array: return
            if get_shift_key_state(): return
            if get_ctrl_key_state():
                cmap_adjust( event )
                return

            cur_xlim = ax.get_xlim()
            cur_ylim = ax.get_ylim()

            xdata = event.xdata # get event x location
            ydata = event.ydata # get event y location

            if event.button == 'up':
                # deal with zoom in
                scale_factor = 1 / base_scale
            elif event.button == 'down':
                # deal with zoom out
                scale_factor = base_scale
            else:
                # deal with something that should never happen
                scale_factor = 1
                print( 'unexpected: event.button=', event.button )

            if scale_factor > 1:
                if self.accumulated_scale > 1.0 - 0.001:
                    self.no_effect_count += 1
                    # print( 'self.no_effect_count=', self.no_effect_count )
                    if self.no_effect_count >= 2:
                        # 元のスケールと位置に戻す。
                        ax.set_xlim( self.orig_xlim )
                        ax.set_ylim( self.orig_ylim )
                        ax.figure.canvas.draw()
                        self.no_effect_count = 0
                    return
            else:
                if self.accumulated_scale < 1.0/256 + 0.001:
                    return

            self.no_effect_count = 0
            self.accumulated_scale *= scale_factor

            new_width = (cur_xlim[1] - cur_xlim[0]) * scale_factor
            new_height = (cur_ylim[1] - cur_ylim[0]) * scale_factor

            relx = (cur_xlim[1] - xdata)/(cur_xlim[1] - cur_xlim[0])
            rely = (cur_ylim[1] - ydata)/(cur_ylim[1] - cur_ylim[0])

            ax.set_xlim([xdata - new_width * (1-relx), xdata + new_width * (relx)])
            ax.set_ylim([ydata - new_height * (1-rely), ydata + new_height * (rely)])
            ax.figure.canvas.draw()

        # 次により、イベントに伴う呼び出しは、
        # それぞれの ZoomPan インスタンスに対して行われている様子。
        self.canvas.mpl_connect('scroll_event', zoom)

        return zoom

    def pan_factory(self, ax_array):
        ax = ax_array[0]

        def onPress(event):
            if get_shift_key_state(): return
            if get_ctrl_key_state(): return
            if event.inaxes != ax: return
            self.cur_xlim = ax.get_xlim()
            self.cur_ylim = ax.get_ylim()
            self.press = self.x0, self.y0, event.xdata, event.ydata
            self.x0, self.y0, self.xpress, self.ypress = self.press

        def onRelease(event):
            if get_shift_key_state(): return
            if get_ctrl_key_state(): return
            self.press = None
            ax.figure.canvas.draw()

        def onMotion(event):
            if get_shift_key_state(): return
            if get_ctrl_key_state(): return
            if self.press is None: return
            if event.inaxes not in ax_array: return

            dx = event.xdata - self.xpress
            dy = event.ydata - self.ypress
            self.cur_xlim -= dx
            self.cur_ylim -= dy

            for ax_ in ax_array:
                ax_.set_xlim(self.cur_xlim)
                ax_.set_ylim(self.cur_ylim)
                ax_.figure.canvas.draw()

        # 下記により、イベントに伴う呼び出しは、
        # それぞれの ZoomPan インスタンスに対して行われている様子。
        self.canvas.mpl_connect('button_press_event',onPress)
        self.canvas.mpl_connect('button_release_event',onRelease)
        self.canvas.mpl_connect('motion_notify_event',onMotion)

        #return the function
        return onMotion



"""
    LrfInfoProxy.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

class LrfInfoProxy:
    def __init__(self, A_data, need_bq=False, B_ratio=None, data=None, bq_bounds=None):
        self.boundary_j = None
        self.sg = sg = SimpleGuinier(A_data)
        self.need_bq_ = need_bq
        self.B_ratio = B_ratio
        self.Rg = sg.Rg
        self.basic_quality = sg.basic_quality
        self.data = data
        self.bq_bounds = bq_bounds

    def need_bq(self):
        return self.need_bq_



"""
    ParamSetUtils.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

def construct_params_type_from_code(func_code, n):
    if func_code < "F0500":
        from ModelParams.EghParams import construct_egh_params_type
        params_type = construct_egh_params_type(n)
    elif func_code < "F0600":
        from ModelParams.LjEghParams import LjEghParams
        params_type = LjEghParams(n)
    elif func_code < "F0700":
        from ModelParams.FdEmgParams import FdEmgParams
        params_type = FdEmgParams(n)
    elif func_code < "F1000":
        from ModelParams.RtEmgParams import RtEmgParams
        params_type = RtEmgParams(n)
    else:
        from ModelParams.StcParams import construct_stochastic_params_type
        params_type = construct_stochastic_params_type(n)
    return params_type



"""
    ElutionDecomposer.py

    Copyright (c) 2018-2023, SAXS Team, KEK-PF
"""
import time
import numpy as np
from scipy.optimize import minimize
import logging
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry, Struct
from molass_legacy.KekLib.NumpyUtils import np_savetxt
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.Models.ElutionCurveModels import (EGHA, EMG, EMGA, model_debug_plot)
from CanvasDialog import CanvasDialog
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
from molass_legacy._MOLASS.SerialSettings import get_setting
from FitRecord import FitRecord, sort_fit_recs
from ModelEvaluator import ModelEvaluator

DEBUG   = False
FIT_DEBUG = False

USE_TILTED_MODEL        = False
USE_PENALTY_PEAK_ERROR  = False
USE_LOGARITHMIC_RATIO   = False
LOGARITHMIC_RATIO       = 0.5
VERY_SMALL_VALUE        = 1e-5
VERY_LARGE_VALUE        = 1e10
RESID_FIND_RATIO        = 0.02  # < 0.03 for the 1st resid peak in Ald
MAX_RESID_PEAK_ONLY     = False

if USE_PENALTY_PEAK_ERROR:
    SCALE_PEAK_ERROR    = 1e-10
else:
    # SCALE_PEAK_ERROR    = 1e3     # ok
    SCALE_PEAK_ERROR    = 50

def proof_plot( decomposer, parent=None ):
    dialog = CanvasDialog( "proof_plot", parent=parent )

    def result_plot( fig ):
        ax  = fig.add_subplot(111)
        ax.plot( decomposer.x, decomposer.y, color='orange' )

        x   = decomposer.x
        residual = decomposer.y.copy()

        for _, func, _ in decomposer.fit_recs:
            ey = func( x )
            ax.plot( ey )
            residual -= ey

        ax.plot( residual )
        fig.tight_layout()

    dialog.show( result_plot, toolbar=True )

class ElutionDecomposer:
    def __init__( self, xcurve, x, y, peaks=None, model=None, d_curve=None, retry_valley=False, deeply=False, hints_dict=None, print_peaks=False, debug=False ):
        self.logger  = logging.getLogger( __name__ )

        if hints_dict is not None:
            self.logger.info('decomposition requested with hints_dict %s.' % str(hints_dict) )

        self.hints_dict = hints_dict

        if peaks is None:
            peaks = xcurve.get_emg_peaks()

        if print_peaks:
            print( 'ElutionDecomposer.__init__: peaks=', peaks )

        self.x  = x
        self.y  = y
        self.max_y = np.max(y)

        default_model = EGHA()
        if model is None:
            model   = default_model

        self.decomp_from_separation = get_setting('decomp_from_separation')

        try:
            self.try_decompose( model, xcurve, peaks, x, y, d_curve, retry_valley, deeply )
            self.logger.info('decomposition resulted into %d records with decomp_from_separation=%d.', len(self.fit_recs), self.decomp_from_separation)
        except:
            etb = ExceptionTracebacker()
            print( etb )
            self.logger.warning( "%s.fit failed. resorting to default_model(EGHA)." % ( model.name ) )
            model = default_model
            self.try_decompose( model, xcurve, peaks, x, y, d_curve, retry_valley, deeply )
            self.logger.warning( "%s.fit done." % ( model.name ) )

        self.result_model_name = model.get_name()

        if debug:
            from OptRecsUtils import debug_plot_opt_recs
            debug_plot_opt_recs(xcurve, self.fit_recs, title="ElutionDecomposer Result")

    def rectify_peaks(self, xcurve, peaks, d_curve):
        rectified = False
        diff_top_x = d_curve.primary_peak_i
        if diff_top_x is None:
            return peaks, rectified

        debug = False

        if debug:
            from molass_legacy.Elution.CurveUtils import simple_plot
            fig = plt.figure()
            ax = fig.gca()
            simple_plot(ax, d_curve)

        for k, peak in enumerate(peaks):
            y = peak.get_model_y(self.x)
            if debug:
                ax.plot(y)
            spx = peak.sigma_points
            if spx[0] < diff_top_x and diff_top_x < spx[1]:
                if debug:
                    ax.plot(diff_top_x, d_curve.spline(diff_top_x), 'o', color='cyan')
                    for x_ in spx:
                        y_ = peak.get_model_y([x_])
                        ax.plot(x_, y_[0], 'o', color='yellow')
                peaks = self.add_peak_at_diff_top_x(xcurve, peaks, k, diff_top_x)
                rectified = True
                break
        if debug:
            fig.tight_layout()
            plt.show()

        return peaks, rectified

    def add_peak_at_diff_top_x(self, xcurve, peaks, k, diff_top_x):
        # task: remove this

        from molass_legacy.ElutionCurveDivider import egha_divide

        d_peaks = egha_divide(xcurve, peaks, k, diff_top_x)
        peaks[k] = d_peaks[0]
        peaks.insert(k+1, d_peaks[1])
        self.logger.warning("one peak has been separated into two peaks from the separation between UV and Xray elutions.")

        return peaks

    def try_decompose( self, model, xcurve, peaks, x, y, d_curve, retry_valley, deeply, debug=False):
        if d_curve is None or not self.decomp_from_separation:
            rectified = False
        else:
            peaks, rectified = self.rectify_peaks(xcurve, peaks, d_curve)

        if rectified:
            self.set_fit_res_just_from_peaks(peaks)
            return

        init_fit_recs = self.compute_fit_recs( model, peaks, x, y, debug=debug)
        if FIT_DEBUG:
            self.debug_plot( x, y, init_fit_recs, title="init_fit_recs with " + model.name )

        if retry_valley:
            resid_y = self.compute_residual_curve( init_fit_recs, x, y, debug=debug )
            try:
                valleys = self.get_valleys_to_modify( xcurve, resid_y )
            except:
                self.logger.warning( "failed to get_valleys_to_modify" )
                valleys = []

            if len(valleys) > 0:
                # TODO: optimize this ratio
                def obj_func( ratios ):
                    y_ = self.modify_y( valleys, resid_y, y, ratios )
                    temp_fit_recs = self.compute_fit_recs( model, peaks, x, y_ )    # usually, do not debug this call to avoid many repeats.
                    temp_resid_y = self.compute_residual_curve( temp_fit_recs, x, y )
                    error = np.sum( [ np.sum( temp_resid_y[slice_]**2 ) for slice_ in valleys ] )
                    return error

                init_ratios = np.ones( len(valleys) )
                result = minimize( obj_func, init_ratios )
                opt_ratios = result.x

                y_ = self.modify_y( valleys, resid_y, y, opt_ratios )
                opt_fit_recs = self.compute_fit_recs( model, peaks, x, y_, debug=debug)
                if False:
                    self.debug_plot( x, y_, opt_fit_recs, title="modified y" )
                    self.debug_plot( x, y, opt_fit_recs, title="original y" )
                self.init_fit_recs = opt_fit_recs
            else:
                self.init_fit_recs = init_fit_recs
        else:
            self.init_fit_recs = init_fit_recs

        if deeply:
            fit_recs = self.add_other_peaks( model, peaks, x, y, self.init_fit_recs )
            self.fit_recs = fit_recs
        else:
            self.fit_recs = self.init_fit_recs

        if FIT_DEBUG:
            self.debug_plot( x, y, self.fit_recs, title="fit_recs with " + model.name )

    def set_fit_res_just_from_peaks(self, peaks):
        fit_recs = []
        for kno, peak in enumerate(peaks):
            # chisqr_n = out.chisqr/self.max_y/(stop-start) # TODO
            chisqr_n = 0
            fit_rec = FitRecord(kno, ModelEvaluator(peak.model, peak.opt_params, sign=peak.sign), chisqr_n, peak)
            fit_recs.append( fit_rec )
        self.init_fit_recs = fit_recs
        self.fit_recs = fit_recs

    def compute_residual_curve( self, fit_recs, x, y, debug=False ):
        resid_y = y.copy()
        for rec in fit_recs:
            func = rec[1]
            resid_y -= func( x )

        if debug:
            self.debug_plot( x, y, fit_recs )

        return resid_y

    def debug_plot( self, x, y, fit_recs, title=None ):
        plt.push()
        fig = plt.figure()
        ax  = fig.add_subplot(111)
        ax.cla()
        if title is not None:
            ax.set_title( title )
        ax.plot( x, y, label='input' )
        resid_y = y.copy()
        for k, rec in enumerate(fit_recs):
            func= rec[1]
            fy = func(x)
            ax.plot( x, fy, label='fit-%d' % k )
            resid_y -= fy
        ax.plot( x, resid_y, label='residuals' )
        ax.legend()
        plt.show()
        plt.pop()

    def compute_residual_error( self, fit_recs=None ):
        if fit_recs is None:
            fit_recs = self.fit_recs
        resid_y = self.compute_residual_curve( fit_recs, self.x, self.y )
        return np.sum( resid_y**2 )

    def get_valleys_to_modify( self, xcurve, resid_y ):
        very_small_abs = xcurve.max_y * 0.01
        abs_resid_y = np.abs( resid_y )

        valleys = []
        for j, info in enumerate( xcurve.peak_info ):
            if j > 0:
                k = j - 1
                b = xcurve.boundaries[k]
                # print( [k], 'b, abs_resid_y[b]=', b, abs_resid_y[b] )
                if abs_resid_y[b] > -xcurve.max_y * 0.05:
                    # print( 'need modify ', b )
                    w   = np.where( abs_resid_y[0:b] < very_small_abs )[0]
                    # print( 'w=', w[-5:] )
                    b_lower = w[-1]
                    w   = b + np.where( abs_resid_y[b:] < very_small_abs )[0]
                    # print( 'w=', w[0:5] )
                    b_upper = w[0]
                    # print( 'b_lower, b_upper=', b_lower, b_upper )
                    valleys.append( slice( b_lower, b_upper ) )
        return valleys

    def modify_y( self, valleys, resid_y, y, ratios ):
        y_ = y.copy()

        for k, slice_ in enumerate(valleys):
            y_[slice_] += resid_y[slice_] * ratios[k]

        return y_

    def compute_fit_recs( self, model, peaks, x, arg_y, k_sign=1, simple_try=True, return_chisqr=False, debug=False ):

        y = arg_y.copy()

        fit_recs = []

        ret_chisqr = 0

        # try fit from the largest to the smallest of peak.top_y
        for k, peak in sorted([(k, rec) for k, rec in enumerate(peaks)], key=lambda p: -p[1].top_y):
            # print([k], peak.top_y)
            f, t    = peak.get_fit_limits()
            start   = f
            stop    = t + 1
            slice_  = slice( start, stop )
            x_  = x[slice_]
            y_  = y[slice_]

            if self.hints_dict is None:
                tau_hint = None
            else:
                xkey = peak.get_xkey()
                tau_hint = self.hints_dict.get(xkey)
                # self.logger.info( '[%d] %s tau_hint=%s' % (k, xkey, str(tau_hint)) )

            params  = model.guess(y_, x=x_, tau_hint=tau_hint)

            if simple_try:
                # out = model.fit(y_, params, x=x_, method='least_squares')
                try:
                    out = model.fit(y_, params, x=x_ )
                except:
                    # self.logger.warning( "%s.fit failed. the %dth element has been ignored." % ( model.name, k ) )
                    # logging is not appropriate in an optimizer loop
                    # 20180206 case, which causes this exception, should be investigated
                    etb = ExceptionTracebacker()
                    print( str(etb) )
                    if debug:
                        try:
                            model_debug_plot( model, out.params,  x_, y_, x, arg_y, tau_hint, "after fit", residual=return_chisqr, before_params=params, print_callstack=True )
                        except:
                            etb = ExceptionTracebacker()
                            print( str(etb) )
                    continue
            else:
                try:
                    out = model.fit(y_, params, x=x_)
                except:
                    etb = ExceptionTracebacker()
                    print( str(etb) )
                    # EMG model sometimes raises ValueError("The input contains nan values")
                    self.logger.warning( "%s.fit default method failed. resorting to 'least_squares'." % ( model.name ) )
                    out = model.fit(y_, params, x=x_, method='least_squares')
                    if debug:
                        try:
                            model_debug_plot( model, out.params,  x_, y_, x, arg_y, tau_hint, "after fit", residual=return_chisqr, before_params=params )
                        except:
                            etb = ExceptionTracebacker()
                            print( str(etb) )

            y -= model.eval(out.params, x=x)

            if debug:
                try:
                    model_debug_plot(model, out.params,  x_, y_, x, arg_y, tau_hint, "after fit", residual=return_chisqr, before_params=params, work_y=y)
                except:
                    etb = ExceptionTracebacker()
                    print( str(etb) )

            # print( [k], out.params )
            kno = k*k_sign if k_sign > 0 else (k+1)*k_sign
            chisqr_n = out.chisqr/self.max_y/(stop-start)

            ret_chisqr += chisqr_n

            # ModelEvaluator is too often constructed. why?
            fit_rec = FitRecord(kno, ModelEvaluator(model, out.params, sign=peak.sign), chisqr_n, peak)
            fit_recs.append( fit_rec )

        ret_fit_recs = sort_fit_recs(fit_recs)
        if return_chisqr:
            return ret_fit_recs, ret_chisqr
        else:
            return ret_fit_recs

    def add_other_peaks( self, model, peaks, x, y, init_fit_recs ):
        resid_y = self.y.copy()

        fit_recs = []

        for k, rec in enumerate( init_fit_recs ):
            func = rec[1]
            fit_recs.append(rec)
            ey = func( self.x )
            resid_y -= ey

        """
        exclude cases where resid_y becomes all NaNs as in 0201006_1
        """
        assert not np.isnan(resid_y[0])

        resid_max_y = np.max( resid_y )
        find_max_y = self.max_y * RESID_FIND_RATIO
        if resid_max_y < find_max_y:
            print('find_giveup_ratio=', find_max_y/self.max_y)
            return init_fit_recs

        try:
            debug_plot = False
            abs_resid_y = np.abs(resid_y)
            resid_curve = ElutionCurve( abs_resid_y, low_quality=True, delay_emg_peaks=True, debug_plot=debug_plot )

            max_h   = None
            max_peak = None
            resid_peaks = resid_curve.get_emg_peaks(max_y=find_max_y, allow_wider=True, orig_y=resid_y, debug=False )

            if False:
                plt.push()
                ax = plt.gca()
                ax.set_title( "Peak recognition in the residual curve" )
                ax.plot( x, resid_y, color='brown', alpha=0.3 )
                ax.plot( x, resid_curve.spline(x), color='green' )
                ax.plot( x[[0,-1]], [find_max_y, find_max_y], ':', color='yellow' )
                for info in resid_curve.peak_info:
                    top_x = info[1]
                    ax.plot( top_x, resid_curve.spline(top_x), 'o', color='yellow' )
                for peak in resid_peaks:
                    top_x = peak.top_x
                    ax.plot( top_x, resid_curve.spline(top_x), 'o', color='red' )
                plt.show()
                plt.pop()

            if MAX_RESID_PEAK_ONLY:
                for peak in resid_peaks:
                    h   = resid_curve.spline( peak.top_x )
                    if max_h is None or h > max_h:
                        max_h = h
                        max_peak = peak

            if len(resid_peaks) > 0:
                # set simple_try=False for the residual curve of May11
                peaks_ = [max_peak] if MAX_RESID_PEAK_ONLY else resid_peaks
                added_fit_recs, ret_chisqr = self.compute_fit_recs( model, peaks_, x, abs_resid_y, k_sign=-1, simple_try=False, return_chisqr=True )
                ry = self.compute_residual_curve( added_fit_recs, x, resid_y )
                chisqr = np.sum( ry**2 )
                # TODO: why ret_chisqr is so small even when chisqr is very large?
                if chisqr < VERY_LARGE_VALUE:
                    if FIT_DEBUG:
                        self.debug_plot( x, resid_y, added_fit_recs, title="added_fit_recs with " + model.name)

                    fit_recs += added_fit_recs

        except Exception as exc:
            etb = ExceptionTracebacker()
            # print(etb)
            self.logger.warning( '%s occurred in ElutionDecomposer.add_other_peaks' % etb.last_line() )

            if False:
                # print( etb )
                plt.plot( x, resid_y )
                plt.plot(  max_peak.top_x, max_peak.top_y, 'o', color='red' )
                for x in [ max_peak.flimL, max_peak.flimR ]:
                    plt.plot(  x, resid_y[int(x)], 'o', color='yellow' )
                plt.show()

        return sort_fit_recs(fit_recs)

    def debug_plot_fitted( self, x, y, fit_rec ):
        plt.plot( x, y, color='orange' )
        func = fit_rec[1]
        plt.plot( x, func(x) )
        plt.show()

    def determine_intervals_deprecated( self, peak_info, x, fit_recs ):

        intervals = []
        for k, info in enumerate( peak_info ):
            x_  = x[0:info[2]+1]
            y_  = fit_recs[k]( x_ )
            n   = np.where( y_ < 1e-6 )[0]
            start = n[-1]
            x_  = x[info[0]:]
            y_  = fit_recs[0][k]( x_ )
            n   = np.where( y_ < 1e-6 )[0]
            stop = info[0] + n[0] + 1
            intervals.append( [start, stop] )
        return intervals

    def decompose_toAB( self ):
        Y_list = []
        C_list = []
        for _, func, _ in self.fit_recs:
            Y = func( self.x )
            Y_list.append( Y )
            C_ = Y / np.max( Y )
            C_list.append( C_ )
            C_list.append( C_**2 )

        if True:
            C_list.append( np.ones( len(self.x) ) )
        else:
            Ye = self.y - Y_list[0] - Y_list[1]
            C_list.append( Ye/np.max( Ye ) )
        self.C_matrix = np.vstack( C_list ).T
        C_pinv = np.linalg.pinv( self.C_matrix )
        return np.dot( C_pinv, self.data )

    def save_components( self, folders, entire_elution=True ):
        assert len(folders) == len(self.fit_recs)
        assert entire_elution

        clear_dirs_with_retry( folders  )

        evector = np.zeros( len(self.qvector) )

        for k in range( len(self.fit_recs) ):
            if not entire_elution and k < 2:
                start, stop = self.intervals[k]
            else:
                start, stop = 0, len(self.y)
            folder  = folders[k]
            # for i in range( 0, self.data.shape[0] ):
            k_slice = slice( k*2, k*2 + 2 )
            for i in range( start, stop ):
                curve = np.dot( self.C_matrix[i,k_slice], self.AB[k_slice,:] ).flatten()
                file = folder + '/SCATTER_%05d.dat' % i
                np_savetxt( file, np.vstack( [ self.qvector, curve, evector ] ).T )



"""
    Optimizer.NaviFrame.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.V2PropOptimizer.PropOptMenu import PropOptMenu

class NaviFrame(Tk.Frame):
    def __init__(self, parent, canvas, arrows_only=False):
        self.parent = parent
        self.canvas = canvas
        Tk.Frame.__init__(self, parent)
        w = Tk.Frame(self)
        w.pack(side=Tk.LEFT)
        w = Tk.Button(self, text="|◀", width=5, command=canvas.get_first)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="◀◀", width=5, command=canvas.get_previous_best)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="◀", width=5, command=canvas.get_previous)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="Best", width=8, command=canvas.get_best)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="▶", width=5, command=canvas.get_next)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="▶▶", width=5, command=canvas.get_next_best)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="▶|", width=5, command=canvas.get_last)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        if arrows_only:
            return

        w = Tk.Button(self, text="Show Parameters", width=14, command=canvas.show_params)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="Complementary View", width=18, command=canvas.show_complementary_view)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = Tk.Button(self, text="MW Ratios", width=12, command=canvas.show_mw_integrity)
        w.pack(side=Tk.LEFT, padx=10, pady=5)
        w = PropOptMenu(self, canvas)
        w.pack(side=Tk.LEFT, padx=10, pady=5)



"""

    ファイル名：   PilatusUtilsNewStyle.py

    処理内容：

       Pilatus 画像データ関連の共通処理

    Copyright (c) 2015-2025, SAXS Team, KEK-PF
"""
import os
import re
from PilatusUtilsOldStyle import get_prefix_info_list, get_dict_info

file_name_re = re.compile(r'(\w+)_(d\d)_(\d{5})\.(\w{3})')

def get_data_info( folder_info, log_file, mask_file,
                    adj_folder, syn_folder, pilatus_counter, counter_id,
                    log_file_path,
                    sample_complete=False, for_test_data=False,
                    logger=None, debug=False ):

    in_folder = folder_info.path
    counter_dict, org_file_dict, adj_file_dict, syn_file_dict = get_dict_info( in_folder, adj_folder, syn_folder, pilatus_counter, counter_id )
    # print('counter_dict=', counter_dict)

    image_files = folder_info.image_files

    # print('len(image_files)=', len(image_files))
    prefix_info_list, text_dict = get_prefix_info_list( in_folder, log_file_path )
    prefix_dict = {}
    for info in prefix_info_list:
        # print(info[0])
        prefix_dict[info[0]] = info

    single_image = folder_info.single_image

    exec_dict = {}
    num_chages_dict = {}

    def get_prefix_info(f):
        parts = f.split('_')
        info = None
        while info is None:
            f_ = '_'.join(parts)
            info = prefix_dict.get(f_)
            if info is not None or len(parts) == 1:
                break
            parts.pop(-1)
        return info

    for k, path in enumerate(sorted(image_files)):
        folder, file= os.path.split(path)
        m = file_name_re.match(file)
        if not m:
            continue

        # print([k], file)
        f1 = m.group(1)
        f2 = m.group(2)
        f3 = m.group(3)

        prefix_info = get_prefix_info(f1)
        if prefix_info is None:
            continue

        num_changes = prefix_info[2]
        exec_date = prefix_info[-1]
        data_dict = exec_dict.get(exec_date)
        if data_dict is None:
            exec_dict[exec_date] = data_dict = {}

        if single_image:
            key = f1
        else:
            key = '_'.join([f1,f3])
        num_chages_dict[key] = num_changes
        rec = data_dict.get(key)
        if rec is None:
            data_dict[key] = rec = []

        pos_info = prefix_info[3]
        if len(pos_info) == 0:
            print("skipped %s, which has no position change info." % file)
            continue

        fkey = '_'.join([f1,f2,f3])
        count = counter_dict.get(fkey)
        if count is None:
            if logger is None:
                import logging
                logger = logging.getLogger(__name__)
            logger.error("No counter info for %s.", fkey)
            continue

        orig_key = re.sub(r'_d\d_', '_d0_', fkey)
        orig_count = counter_dict.get(orig_key)
        ratio = count/orig_count

        if debug:
            if k < 12:
                print([k], fkey, orig_count, count, ratio)

        adj_file = adj_file_dict.get( key )
        syn_file = syn_file_dict.get( key )
        i = len(rec)
        rec.append([file, pos_info[i], ratio, adj_file, syn_file])

    def alphanum_key(pair):
        ret_key = [int(c) if c.isdecimal() else c for c in re.split('([0-9]+)', pair[0])]
        return ret_key, pair[1]

    data_array = []
    for edate, data_dict in sorted(exec_dict.items()):
        for key, value in sorted(data_dict.items(), key=alphanum_key):
            data_array.append([key, value, num_chages_dict[key]])

    if for_test_data:
        assert False
    else:
        return log_file, mask_file, data_array, pilatus_counter



"""
    Solvers.ABC.ParameterUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from pyabc import Parameter

PARAM_KEY_FORMAT = "p%02d"

def parameter_to_vector(parameter):
    return np.array([parameter[PARAM_KEY_FORMAT % k] for k in range(len(parameter))])

def vector_to_parameter(values):
    pdict = {}
    for i, v in enumerate(values):
        pdict[PARAM_KEY_FORMAT % i] = v
    return Parameter(pdict)